(*
Copyright 2009-2013 René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Mini_Algebraic_Numbers
imports
  "~~/src/HOL/NthRoot"
  "$AFP/Sqrt_Babylonian/Sqrt_Babylonian"
  "$AFP/Abstract-Rewriting/SN_Orders"
  Show_Instances
begin

text {* In this theory we consider all numbers of the form $a + b\sqrt{p}$ 
  for some fixed positive p where $sqrt(p)$ is irrational and where $a$ and $b$ range 
  over the rational numbers. These numbers are a subset of the
  algebraic numbers which we call mini algebraic numbers. It is easy to see that
  the mini algebraic numbers are a strict superset of the rational numbers.

  We show that the mini algebraic numbers forms a linear ordered field,
  where we implement the field operations directly on pairs $(a,b)$.
  To define the orders, we instead use an embedding into the real numbers
  and additionaly characterize both the strict and the non-strict order
  via an executable expression. 

  Furthermore, we show that mini algebraic numbers can serve as a carrier for 
  polynomial orders. *}

text {* Here we fix p *}
definition p where "p \<equiv> 2 :: rat" 

lemma sqrt_p_irrat: "r * r = p \<Longrightarrow> False" 
proof -
  have "set (sqrt_rat p) = {}" unfolding p_def by eval
  thus "r * r = p \<Longrightarrow> False" by auto
qed

lemma p_gt_0[simp]: "p > 0" unfolding p_def by simp

hide_fact p_def 
text {* And now we forget which p was chosen. *}

lemma p0[simp]: "p \<ge> 0" using p_gt_0 by arith
lemma ratp0[simp]: "real_of_rat p \<ge> 0" using p0 by (metis of_rat_0 of_rat_less_eq)

datatype mini_alg_number = Sp rat rat

text {* The number @{term "Sp a b"} represents $a + b \sqrt{p}$. *}

text {* The set of all these numbers forms a field *}

instantiation mini_alg_number :: field
begin
definition zero_mini_alg_number where [simp]: "zero_mini_alg_number = Sp 0 0"
fun plus_mini_alg_number where "plus_mini_alg_number (Sp c1 d1) (Sp c2 d2) = Sp (c1 + c2) (d1 + d2)"
fun minus_mini_alg_number where "minus_mini_alg_number (Sp c1 d1) (Sp c2 d2) = Sp (c1 - c2) (d1 - d2)"
fun uminus_mini_alg_number where "uminus_mini_alg_number (Sp c1 d1) = Sp (- c1) (- d1)"
fun times_mini_alg_number where "times_mini_alg_number (Sp c1 d1) (Sp c2 d2) = Sp (c1 * c2 + p * d1 * d2) (c1 * d2 + c2 * d1)"
definition one_mini_alg_number where [simp]: "one_mini_alg_number = Sp 1 0"
abbreviation inverse_sp where "inverse_sp a b \<equiv> let d = inverse (a * a - p * b * b) in Sp (a * d) (- b * d)"
fun inverse_mini_alg_number :: "mini_alg_number \<Rightarrow> mini_alg_number" where
  "inverse_mini_alg_number (Sp a b) = inverse_sp a b"
fun divide_mini_alg_number where "divide_mini_alg_number x (Sp a b) = x * inverse_sp a b"
declare field_simps[simp]
instance
proof
  fix a b c :: mini_alg_number
  show "a + b = b + a" by (cases a, cases b, auto)
  show "0 + a = a" by (cases a, auto)
  show "a + b + c = a + (b + c)" by (cases a, cases b, cases c, auto)
  show "- a + a = 0" by (cases a, auto)
  show "a - b = a + - b" by (cases a, cases b, auto)
  show "a * b * c = a * (b * c)" by (cases a, cases b, cases c, auto)
  show "a * b = b * a" by (cases a, cases b, auto)
  show "(a + b) * c = a * c + b * c" by (cases a, cases b, cases c, auto)
  show "1 * a = a" by (cases a, auto)
  show "a / b = a * inverse b" by (cases b, auto)
next
  fix x :: mini_alg_number
  assume 0: "x \<noteq> 0"
  obtain a b where x: "x = Sp a b" by (cases x, auto)
  from x 0 have one: "a \<noteq> 0 \<or> b \<noteq> 0" by simp
  def d \<equiv> "a * a - p * b * b"
  have "inverse x * x = Sp (d * inverse d) 0" unfolding x d_def 
    by (simp add: Let_def)
  also have "d * inverse d = 1" 
  proof (rule right_inverse)
    show "d \<noteq> 0"
    proof
      assume "d = 0"
      hence id: "a * a - p * b * b = 0" unfolding d_def .
      with one have b: "b \<noteq> 0" by (cases "b = 0", auto)
      from id b have "(a / b) * (a / b) = p" by simp
      from sqrt_p_irrat[OF this] 
      show False by simp
    qed        
  qed
  finally show "inverse x * x = 1" by simp
qed auto
end

text {* for defining the order on mini_alg_numbers, we use the order on reals and
  embed the mini_alg_numbers into the reals *}

fun real_of :: "mini_alg_number \<Rightarrow> real" where
  "real_of (Sp a b) = of_rat a + of_rat b * sqrt (of_rat p)"

lemma real_of_diff[simp]: "real_of (x - y) = real_of x - real_of y"
  by (cases x, cases y, simp add: of_rat_diff)

lemma real_of_minus[simp]: "real_of ( - y) = - real_of y"
  by (cases y, simp add: of_rat_minus)

lemma real_of_add[simp]: "real_of (x + y) = real_of x + real_of y"
  by (cases x, cases y, simp add: of_rat_add)

lemma sqrt: "sqrt (of_rat p) * sqrt (of_rat p) = of_rat p"
  by (simp add: comm_semiring_1_class.normalizing_semiring_rules)

lemma real_of_mult[simp]: "real_of (x * y) = real_of x * real_of y"
  by (cases x, cases y, simp add: of_rat_mult of_rat_add sqrt)

lemma real_of_inj: assumes "real_of x = real_of y" shows "x = y"
proof -
  obtain c1 d1 where x: "x = Sp c1 d1" by (cases x, auto)
  obtain c2 d2 where y: "y = Sp c2 d2" by (cases y, auto)
  def c \<equiv> "c2 - c1"
  def d \<equiv> "d1 - d2"
  from assms[unfolded x y]
  have id: "of_rat c1 + of_rat d1 * sqrt (of_rat p) = of_rat c2 + of_rat d2 * sqrt (of_rat p)" by simp
  hence id: "of_rat d * sqrt (of_rat p) = of_rat c" unfolding c_def d_def of_rat_diff
    by simp
  have c: "c = 0" 
  proof (rule ccontr)
    assume c: "c \<noteq> 0"
    from id have "of_rat (inverse c * d) * sqrt (of_rat p) = of_rat (inverse c * c)" unfolding of_rat_mult by simp
    also have "of_rat (inverse c * c) = 1" using c by auto
    finally obtain e where id: "of_rat e * sqrt (of_rat p) = 1" by blast
    hence e: "e \<noteq> 0" by auto
    from id have "of_rat (inverse e * e) * sqrt (of_rat p) = of_rat (inverse e)" unfolding of_rat_mult by auto
    with e have "sqrt (of_rat p) = of_rat (inverse e)" by auto
    then obtain f where "sqrt (of_rat p) = of_rat f" by auto
    with arg_cong[OF this, of "\<lambda> x. x ^ 2"] have "of_rat p = of_rat f * of_rat f" 
      by (metis of_rat_eq_iff of_rat_mult sqrt)
    hence "of_rat p = of_rat (f * f)" by (simp add: of_rat_mult)
    hence "f * f = p" by simp
    from sqrt_p_irrat[OF this] show False by simp
  qed
  with id p_gt_0 have d: "d = 0" by simp
  show ?thesis unfolding x y using c d unfolding c_def d_def by auto
qed

corollary unique_0: "real_of x = 0 \<Longrightarrow> x = Sp 0 0"
  using real_of_inj[of x "Sp 0 0"] by auto

fun geq :: "mini_alg_number \<Rightarrow> mini_alg_number \<Rightarrow> bool" where
  "geq x y = (real_of x \<ge> real_of y)"

fun gt :: "mini_alg_number \<Rightarrow> mini_alg_number \<Rightarrow> bool" where
  "gt x y = (real_of x > real_of y)"


instantiation mini_alg_number :: linordered_field
begin
fun less_eq_mini_alg_number where [code_unfold]: "less_eq_mini_alg_number x y = geq y x"
fun less_mini_alg_number where [code_unfold]: "less_mini_alg_number x y = gt y x"
fun sgn_mini_alg_number where "sgn_mini_alg_number x = (if x = 0 then (0 :: mini_alg_number) else (if geq x 0 then 1 else - 1))"
fun abs_mini_alg_number where "abs_mini_alg_number x = (if geq x 0 then x else - x)"
instance
proof
  fix x y :: mini_alg_number
  assume "x \<le> y" and "y \<le> x"
  hence "real_of x = real_of y" by auto
  from real_of_inj[OF this] show "x = y" .
qed (insert unique_0, auto)
end

lemma real_of_max[simp]: "real_of (max x y) = max (real_of x) (real_of y)"
  unfolding max_def by simp

lemma mini_alg_of_nat: "of_nat n = Sp (rat_of_nat n) 0"
  by (induct n, auto)
  
lemma mini_alg_of_int: "of_int x = Sp (rat_of_int x) 0"
proof -
  {
    fix y :: int
    assume y: "y \<ge> 0"
    hence "of_int y = of_int (int (nat y))" by simp
    also have "\<dots> = of_nat (nat y)" unfolding of_int_of_nat_eq ..
    also have "\<dots> = Sp (rat_of_nat (nat y)) 0" unfolding mini_alg_of_nat ..
    also have "\<dots> = Sp (rat_of_int y) 0" using y by auto
    finally have "of_int y = Sp (rat_of_int y) 0" .
  } note main = this
  show ?thesis  
  proof (cases "x \<ge> 0")
    case True
    from main[OF this] show ?thesis .
  next
    case False
    hence x: "- x \<ge> 0" by auto
    from arg_cong[OF main[OF this], of "\<lambda> x. - x"] show ?thesis by simp
  qed
qed
  
lemma real_of_of_nat[simp]: "real_of (of_nat r) = real_of_nat r"
  by (induct r, auto)

lemma real_of_of_int[simp]: "real_of (of_int r) = real_of_int r"
  unfolding of_int_of_nat
  by (cases "r < 0", auto)

instantiation mini_alg_number :: floor_ceiling
begin
definition floor_mini_alg_number :: "mini_alg_number \<Rightarrow> int" where
  "floor_mini_alg_number r = floor (real_of r)"
instance
proof
  fix x :: mini_alg_number
  from ex_le_of_int[of "real_of x"] obtain y where r: "real_of x \<le> real_of_int y" ..
  hence "x \<le> Sp (rat_of_int y) 0" by simp
  also have "\<dots> = of_int y" unfolding mini_alg_of_int ..
  finally show "\<exists>z. x \<le> of_int z" ..
  show "of_int \<lfloor>x\<rfloor> \<le> x \<and> x < of_int (\<lfloor>x\<rfloor> + 1)" unfolding floor_mini_alg_number_def
    using floor_correct[of "real_of x"] by auto
qed
end


instantiation mini_alg_number :: poly_carrier
begin
fun max0_mini_alg_number where "max0_mini_alg_number n = (if geq n 0 then n else 0)"
fun ge_mini_alg_number where [code_unfold]: "ge_mini_alg_number x y = geq x y"
instance 
proof
  fix x y z :: mini_alg_number
  assume x: "x \<succeq> 0" 
  show "y \<succeq> z \<Longrightarrow> x * y \<succeq> x * z" using x  
    using mult_left_mono[of "real_of z" "real_of y" "real_of x"] by simp
next
  fix z x y :: mini_alg_number
  show "z \<succeq> 0 \<Longrightarrow> x \<succeq> y \<Longrightarrow> x * z \<succeq> y * z"
    using mult_right_mono[of "real_of y" "real_of x" "real_of z"] by simp
qed auto
end

instantiation mini_alg_number :: bin_max_ordered_semiring_1
begin
instance 
proof 
  fix y :: mini_alg_number
  from ex_le_of_nat[of y]
  show "\<exists> x. of_nat x \<succeq> y" by simp
qed auto
end
 
instantiation mini_alg_number :: "show" begin
  definition show_mini_alg_number :: "mini_alg_number \<Rightarrow> string"
  where "show_mini_alg_number x \<equiv> let sp = shows ''*sqrt('' \<circ> shows p \<circ> shows '')'' in 
    case x of (Sp r i) \<Rightarrow> 
      if i = 0 then shows r [] else 
      if r = 0 then (shows i \<circ> sp) [] else
      if i < 0 then ((shows ''('' \<circ> shows r \<circ> shows i \<circ> sp \<circ> shows '')'') []) 
      else ((shows ''('' \<circ> shows r \<circ> shows ''+'' \<circ> shows i \<circ> sp \<circ> shows '')'') [])"
  definition "shows_prec (d::nat) x \<equiv> shows_string (show_mini_alg_number x)"

  lemma assoc_mini_alg: "shows_prec d (x::mini_alg_number) r @ s = shows_prec d x (r @ s)" unfolding shows_prec_mini_alg_number_def by  auto
  standard_shows_list assoc_mini_alg
end

definition mini_alg_mono :: "mini_alg_number \<Rightarrow> bool" where 
  "mini_alg_mono x \<equiv> x \<ge> 1"

definition mini_alg_gt :: "mini_alg_number \<Rightarrow> mini_alg_number \<Rightarrow> mini_alg_number \<Rightarrow> bool" where 
  "mini_alg_gt \<delta> x y \<equiv> \<delta> \<le> x - y"

lemma mini_alg_gt_SN: assumes dpos: "\<delta> > 0" shows "SN {(x,y). 0 \<le> y \<and> mini_alg_gt \<delta> x y}"
proof
  fix f
  assume f: "\<forall>i. (f i, f (Suc i)) \<in> {(x, y). 0 \<le> y \<and> mini_alg_gt \<delta> x y}"
  def g \<equiv> "\<lambda> i. real_of (f i)"
  def d \<equiv> "real_of \<delta>"
  from dpos have d:"d > 0" unfolding d_def by simp
  {
    fix i
    note i = spec[OF f, of i]
    from i have 0: "0 \<le> g (Suc i)" unfolding g_def by auto
    from i have diff: "g  i \<ge> g (Suc i) + d" unfolding g_def d_def mini_alg_gt_def by auto
    from 0 diff d have 0: "0 \<le> g i" by arith
    note 0 diff
  } note diff = this
  {
    fix i
    have "g i \<le> g 0 - d * of_nat i"
    proof (induct i)
      case (Suc i)
      thus ?case using diff[of i] by (auto simp: field_simps)
    qed simp
  } note gi = this
  {
    fix r :: real
    have "of_nat (nat (ceiling r)) \<ge> r" 
      by (metis ceiling_le_zero le_of_int_ceiling less_le_not_le nat_0_iff not_less of_nat_0 of_nat_nat)
  } note ceil_elim = this
  def i \<equiv> "nat (ceiling ((g 0 + 1) / d))"
  from gi[of i] have "g i - g 0 \<le> - d * of_nat (nat (ceiling ((g 0 + 1) / d)))" unfolding i_def by simp
  also have "\<dots> \<le> - d * ((g 0 + 1) / d)" using ceil_elim[of "(g 0 + 1) / d"] d 
    by (metis le_imp_neg_le minus_mult_commute mult_le_cancel_left_pos)
  also have "\<dots> = - g 0 - 1" using d by auto
  finally show False using diff(1)[of i] by arith
qed

lemma mini_alg_interpretation: assumes dpos: "\<delta> > 0" and default: "\<delta> \<le> def"
  shows "SN_strict_mono_ordered_semiring_1 def (mini_alg_gt \<delta>) mini_alg_mono"
proof -
  from dpos default have defz: "0 \<le> def" by auto
  show ?thesis
  proof (unfold_locales)
    show "SN {(x,y). y \<succeq> 0 \<and> mini_alg_gt \<delta> x y}" 
      using mini_alg_gt_SN[OF dpos] by simp
  next
    fix x y z :: mini_alg_number
    assume "mini_alg_mono x" and yz: "mini_alg_gt \<delta> y z"
    hence x: "1 \<le> x" unfolding mini_alg_mono_def by simp
    have "\<exists> d > 0. mini_alg_gt \<delta> = (\<lambda> x y. d \<le> x - y)"
      by (rule exI[of _ \<delta>], insert dpos, auto simp: mini_alg_gt_def fun_eq_iff)
    from this obtain d where d: "0 < d" and rat: "mini_alg_gt \<delta> = (\<lambda> x y. d \<le> x - y)" by auto
    from yz have yzd: "d \<le> y - z" by (simp add: rat)
    show "mini_alg_gt \<delta> (x * y) (x * z)"
    proof (simp only: rat)
      let ?p = "(x - 1) * (y - z)"
      from x have x1: "0 \<le> x - 1" by linarith
      from yzd d have yz0: "0 \<le> y - z" by auto
      have "0 \<le> ?p" 
        by (rule mult_nonneg_nonneg[OF x1 yz0])
      have "x * y - x * z = x * (y - z)" using right_diff_distrib[of x y z] by auto
      also have "\<dots> = ((x - 1) + 1) * (y - z)" by auto
      also have "\<dots> = ?p + 1 * ( y - z)" by (rule ring_distribs(2))
      also have "\<dots> = ?p + (y - z)" by (metis comm_monoid_mult_class.mult.left_neutral)
      also have "\<dots> \<succeq> (0 + d)" using yzd `0 \<le> ?p` by auto
      finally 
      show "d \<le> x * y - x * z" by simp
    qed
  next
    fix x y z
    assume "mini_alg_gt \<delta> x y" "mini_alg_gt \<delta> y z"
    thus "mini_alg_gt \<delta> x z" unfolding mini_alg_gt_def using dpos by auto
  qed (insert dpos defz default, auto simp: mini_alg_gt_def)
qed

lemma mini_alg_poly: assumes dpos: "\<delta> > 0" and default: "\<delta> \<le> def"
  shows "poly_order_carrier def (mini_alg_gt \<delta>) (1 \<le> \<delta>) False"
proof -
  from mini_alg_interpretation[OF dpos default] 
  interpret SN_strict_mono_ordered_semiring_1 "def" "mini_alg_gt \<delta>" mini_alg_mono .
  interpret poly_order_carrier "def" "mini_alg_gt \<delta>" False False
  proof(unfold_locales)
    fix y z x :: mini_alg_number
    assume gt: "mini_alg_gt \<delta> y z" and ge: "x \<succeq> 1"
    from ge have ge: "x \<succeq> 0" and m: "mini_alg_mono x" unfolding mini_alg_mono_def by auto
    show "mini_alg_gt \<delta> (y * x) (z * x)"
      using mono[OF m gt ge] by (auto simp: field_simps)
  qed (insert dpos, auto simp: mini_alg_gt_def)
  show ?thesis
  proof(unfold_locales)
    fix x y :: mini_alg_number and n :: nat
    assume one: "1 \<le> \<delta>" and gt: "mini_alg_gt \<delta> x y" and y: "y \<succeq> 0" and n: "1 \<le> n"
    then obtain p where n: "n = Suc p" and x: "x \<succeq> 1" and y2: "0 \<le> y" and xy: "x \<succeq> y" by (cases n, auto simp: mini_alg_gt_def)
    show "mini_alg_gt \<delta> (x ^ n) (y ^ n)" 
    proof (simp only: n, induct p)
      case (Suc p)
      from times_gt_mono[OF this x]
      have one: "mini_alg_gt \<delta> (x ^ Suc (Suc p)) (x * y ^ Suc p)" by (auto simp: field_simps)
      also have "\<dots> \<succeq> y * y ^ Suc p" 
        by (rule times_left_mono[OF _ xy], auto simp: zero_le_power[OF y2, of "Suc p", simplified])
      finally show ?case by auto
    qed (insert gt, auto simp del: one_mini_alg_number_def)
  qed (rule times_gt_mono, simp, simp, 
      rule gt_imp_ge, auto)
qed


lemma mini_alg_minimal_delta: assumes "\<And> x  y :: mini_alg_number. (x,y) \<in> set xys \<Longrightarrow> x > y"
  shows "\<exists> \<delta> > 0. \<forall> x y. (x,y) \<in> set xys \<longrightarrow> mini_alg_gt \<delta> x y"
using assms 
proof (induct xys)
  case Nil
  show ?case by (rule exI[of _ 1], auto)
next
  case (Cons xy xys)
  show ?case
  proof (cases xy)
    case (Pair x y)
    with Cons have "x > y" by auto
    then obtain d1 where "d1 = x - y" and d1pos: "d1 > 0" and "d1 \<le> x - y" 
      by (metis (full_types) le_iff_diff_le_0 not_less order_refl)
    hence xy: "mini_alg_gt d1 x y" unfolding mini_alg_gt_def by auto
    from Cons obtain d2 where d2pos: "d2 > 0" and xys: "\<forall> x y. (x, y) \<in> set xys \<longrightarrow> mini_alg_gt d2 x y" by auto
    obtain d where d: "d = min d1 d2" by auto
    with d1pos d2pos xy have dpos: "d > 0" and "mini_alg_gt d x y" unfolding mini_alg_gt_def by linarith+
    with xys d Pair have "\<forall> x y. (x,y) \<in> set (xy # xys) \<longrightarrow> mini_alg_gt d x y" unfolding mini_alg_gt_def 
      by (metis min_le_iff_disj prod.inject set_ConsD)
    with dpos show ?thesis by auto 
  qed
qed

interpretation weak_mini_alg_SN: weak_SN_strict_mono_ordered_semiring_1 "op >" 1 mini_alg_mono
proof
  fix xysp :: "(mini_alg_number \<times> mini_alg_number) list"
  assume orient: "\<forall> x y. (x,y) \<in> set xysp \<longrightarrow> x > y" 
  obtain xys where xsy: "xys = (1,0) # xysp" by auto
  with orient have "\<And> x y. (x,y) \<in> set xys \<Longrightarrow> x > y" by auto
  with mini_alg_minimal_delta have "\<exists> \<delta> > 0. \<forall> x y. (x,y) \<in> set xys \<longrightarrow> mini_alg_gt \<delta> x y" by auto
  then obtain \<delta> where dpos: "\<delta> > 0" and orient: "\<And> x y. (x,y) \<in> set xys \<Longrightarrow> mini_alg_gt \<delta> x y" by auto
  from orient have orient1: "\<forall> x y. (x,y) \<in> set xysp \<longrightarrow> mini_alg_gt \<delta> x y" and orient2: "mini_alg_gt \<delta> 1 0" unfolding xsy by auto
  from orient2 have oned: "\<delta> \<le> 1" unfolding mini_alg_gt_def by auto
  show " \<exists>gt. SN_strict_mono_ordered_semiring_1 1 gt mini_alg_mono \<and> (\<forall>x y. (x, y) \<in> set xysp \<longrightarrow> gt x y)" 
    by (intro exI conjI, rule mini_alg_interpretation[OF dpos oned], rule orient1)
qed


text {* for gt and geq we need to provide executable algorithms *}

lemma [code]: "geq x y = (case x - y of Sp c d \<Rightarrow> 
  (c \<ge> 0 \<and> c * c \<ge> p * d * d \<or> d \<ge> 0 \<and> p * d * d \<ge> c * c))"
proof -
  have square_lesseq_square: "\<And> x y. 0 \<le> (x :: real) \<Longrightarrow> 0 \<le> y \<Longrightarrow> (x \<le> y) = (x * x \<le> y * y)"
    by (metis eq_iff leD le_neq_trans less_imp_le mult_le_0_iff mult_less_le_imp_less mult_mono mult_zero_right not_leE zero_le_square)
  obtain c d where xy: "Sp c d = x - y" by (cases "x - y", auto)
  have "geq x y \<longleftrightarrow> real_of x - real_of y \<ge> 0"
    unfolding geq.simps by arith
  also have "\<dots> \<longleftrightarrow> real_of (Sp c d) \<ge> 0" unfolding real_of_diff xy ..
  also have "\<dots> \<longleftrightarrow> of_rat c + of_rat d * sqrt (of_rat p) \<ge> 0" by simp
  also have "\<dots> \<longleftrightarrow> of_rat c \<ge> - of_rat d * sqrt (of_rat p)" by arith
  also have "\<dots> \<longleftrightarrow> c \<ge> 0 \<and> c * c \<ge> p * d * d \<or> d \<ge> 0 \<and> p * d * d \<ge> c * c"
  proof (cases "c \<ge> 0")
    case True note c = this
    hence c': "of_rat c \<ge> (0 :: real)" by (metis of_rat_0 of_rat_less_eq)
    show ?thesis
    proof (cases "d \<ge> 0")
      case True note d = this
      hence "- of_rat d * sqrt (of_rat p) \<le> 0" using p0 by (metis minus_mult_commute minus_mult_right neg_le_0_iff_le of_rat_eq_0_iff of_rat_less_eq real_sqrt_eq_zero_cancel_iff real_sqrt_le_iff zero_le_mult_iff)
      from order_trans[OF this c'] have "- of_rat d * sqrt (of_rat p) \<le> of_rat c" .
      thus ?thesis using c d by auto
    next
      case False
      hence d: "d \<le> 0" by auto
      hence d': "of_rat d \<le> (0 :: real)" by (metis of_rat_0 of_rat_less_eq)
      hence pos: "- of_rat d * sqrt (of_rat p) \<ge> 0" using p0 by (metis neg_0_le_iff_le ratp0 real_sqrt_ge_0_iff zero_le_mult_iff)
      have "- of_rat d * sqrt (of_rat p) \<le> of_rat c \<longleftrightarrow> of_rat d * (of_rat d * (sqrt (of_rat p) * sqrt (of_rat p))) \<le> of_rat c * of_rat c"
        unfolding square_lesseq_square[OF pos c'] by simp
      also have "\<dots> \<longleftrightarrow> real_of_rat (d * d * of_rat p) \<le> of_rat (c * c)"
        unfolding sqrt of_rat_mult[symmetric] of_rat_less_eq by simp
      finally have eq: "- of_rat d * sqrt (of_rat p) \<le> of_rat c \<longleftrightarrow> real_of_rat (d * d * of_rat p) \<le> of_rat (c * c)" .
      show ?thesis unfolding eq unfolding of_rat_less_eq using False c eq by simp
    qed
  next
    case False note cFalse = this
    hence c: "c \<le> 0" by auto
    hence c': "- real_of_rat c \<ge> 0" by (metis False less_imp_not_eq2 less_trans neg_less_0_iff_less neg_less_nonneg not_leE of_rat_less of_rat_minus)
    show ?thesis
    proof (cases "d \<ge> 0")
      case False
      from cFalse have  neg: "of_rat c < (0 :: real)" by (metis c' le_neq_trans neg_0_le_iff_le of_rat_eq_0_iff order_refl)
      from False have pos: "- of_rat d * sqrt (of_rat p) \<ge> 0" using p0 by (metis le_numeral_extra(3) mult_zero_left neg_0_le_iff_le neg_less_eq_nonneg of_rat_less_eq of_rat_minus real_sqrt_ge_0_iff zero_le_mult_iff)
      from neg pos have "\<not> (- of_rat d * sqrt (of_rat p) \<le> real_of_rat c)" by arith
      thus ?thesis using False cFalse by auto
    next
      case True
      hence pos: "of_rat d * sqrt (of_rat p) \<ge> 0" using p0 by (metis of_rat_0 of_rat_less_eq real_sqrt_ge_0_iff split_mult_pos_le)
      have "- of_rat d * sqrt (of_rat p) \<le> real_of_rat c \<longleftrightarrow> - of_rat c \<le> of_rat d * sqrt (of_rat p)" by arith
      also have "\<dots> \<longleftrightarrow> real_of_rat (c * c) \<le> of_rat (d * d * (of_rat p))" unfolding square_lesseq_square[OF c' pos]
        by (simp add: sqrt of_rat_mult)
      also have "\<dots> \<longleftrightarrow> c * c \<le> d * d * (of_rat p)" unfolding of_rat_less_eq ..
      finally have eq: "- of_rat d * sqrt (of_rat p) \<le> of_rat c \<longleftrightarrow> c * c \<le> d * d * (of_rat p)" by auto
      show ?thesis unfolding eq using cFalse True by auto
    qed
  qed
  finally show ?thesis unfolding xy[symmetric] by simp
qed

lemma [code]: "gt x y = (geq x y \<and> x \<noteq> y)" unfolding gt.simps geq.simps using real_of_inj[of x y]
  by fastforce

definition mini_of_rat :: "rat \<Rightarrow> mini_alg_number" where
  "mini_of_rat x \<equiv> Sp x 0"

definition sqrtp :: "mini_alg_number" where
  "sqrtp \<equiv> Sp 0 1"

(* hide implementation details *)
lemmas mini_alg_simps = 
  less_eq_mini_alg_number.simps
  less_mini_alg_number.simps
  sgn_mini_alg_number.simps
  abs_mini_alg_number.simps
  max0_mini_alg_number.simps
  ge_mini_alg_number.simps
  zero_mini_alg_number_def
  one_mini_alg_number_def

declare mini_alg_simps[simp del]

hide_const gt
hide_const geq
hide_const Sp

hide_const (open) p

end

