(*
Copyright 2009-2013 René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Non_Inf_Order
imports 
  Trs 
  "$AFP/Abstract-Rewriting/SN_Order_Carrier"
begin

text {* We define non infinitesmal orders and related notions 
  as in the CADE07 "Bounded Increase" paper. *}

text {* non infinitesmal is the same as in CADE07 *}  
definition non_inf :: "'a rel \<Rightarrow> bool"
 where "non_inf r \<equiv> \<forall> a f. \<exists> i. (f i, f (Suc i)) \<notin> r \<or> (f i, a) \<notin> r"

lemma non_infI[intro]: assumes "\<And> a f. \<lbrakk> \<And> i. (f i, f (Suc i)) \<in> r\<rbrakk> \<Longrightarrow> \<exists> i. (f i, a) \<notin> r"
  shows "non_inf r"
  using assms unfolding non_inf_def by blast

lemma non_infE[elim]: assumes "non_inf r" and "\<And> i. (f i, f (Suc i)) \<notin> r \<or> (f i, a) \<notin> r \<Longrightarrow> P"
  shows P
  using assms unfolding non_inf_def by blast

lemma non_inf_image: 
  assumes ni: "non_inf r" and image: "\<And> a b. (a,b) \<in> s \<Longrightarrow> (f a, f b) \<in> r"
  shows "non_inf s"
proof
  fix a g
  assume s: "\<And> i. (g i, g (Suc i)) \<in> s"
  def h \<equiv> "f o g"
  from image[OF s] have h: "\<And> i. (h i, h (Suc i)) \<in> r" unfolding h_def comp_def .
  from non_infE[OF ni, of h] have "\<And> a. \<exists> i. (h i, a) \<notin> r" using h by blast
  thus "\<exists>i. (g i, a) \<notin> s" using image unfolding h_def comp_def by blast
qed
  
lemma non_inf_rat_gt: assumes delta: "\<delta> > 0"
  shows "non_inf {(a,b) . rat_gt \<delta> a b}" (is "non_inf ?r")
proof
  let ?gt = "rat_gt \<delta>"
  fix a :: rat and f
  assume "\<And> i. (f i, f (Suc i)) \<in> ?r"
  hence gt: "\<And> i. ?gt (f i) (f (Suc i))" by simp
  {
    fix i
    have "f i \<le> f 0 - \<delta> * of_nat i"
    proof (induct i)
      case (Suc i)
      thus ?case using gt[of i, unfolded rat_gt_def] by (auto simp: field_simps)
    qed simp
  } note fi = this
  {
    fix r :: rat
    have "of_nat (nat (ceiling r)) \<ge> r" 
      by (metis ceiling_le_zero le_of_int_ceiling less_le_not_le nat_0_iff not_less of_nat_0 of_nat_nat)
  } note ceil_elim = this
  def i \<equiv> "nat (ceiling ((f 0 - a) / \<delta>))"
  from fi[of i] have "f i - f 0 \<le> - \<delta> * of_nat (nat (ceiling ((f 0 - a) / \<delta>)))" unfolding i_def by simp
  also have "\<dots> \<le> - \<delta> * ((f 0 - a) / \<delta>)" using ceil_elim[of "(f 0 - a) / \<delta>"] delta 
    by (metis le_imp_neg_le minus_mult_commute mult_le_cancel_left_pos)
  also have "\<dots> = - f 0 + a" using delta by auto
  also have "\<dots> < - f 0 + a + \<delta>" using delta by auto
  finally have "\<not> ?gt (f i) a" unfolding rat_gt_def by arith
  thus "\<exists> i. (f i, a) \<notin> ?r" by blast
qed

(* TODO: replace rat_gt_SN in SN_Order_Carrier by this shorter proof *)
lemma rat_gt_SN': assumes dpos: "\<delta> > 0" shows "SN {(x,y). 0 \<le> y \<and> rat_gt \<delta> x y}"
proof 
  fix f
  assume "\<forall>i. (f i, f (Suc i)) \<in> {(x, y). 0 \<le> y \<and> rat_gt \<delta> x y}" 
  hence all: "\<And> i. (f i, f (Suc i)) \<in> {(x, y). 0 \<le> y \<and> rat_gt \<delta> x y}" ..
  from non_inf_rat_gt[OF dpos] have ni: "non_inf {(a,b) . rat_gt \<delta> a b}" .
  show False
  proof (rule non_infE[OF ni])
    fix i
    assume "(f i, f (Suc i)) \<notin> {(a, b). rat_gt \<delta> a b} \<or> (f i, - \<delta>) \<notin> {(a, b). rat_gt \<delta> a b}"
    with all[of i] have "\<not> rat_gt \<delta> (f i) (- \<delta>)" by auto
    moreover from all[of "Suc i"] all[of i] have "f i \<ge> 0" using dpos unfolding rat_gt_def by auto
    ultimately show False unfolding rat_gt_def by auto
  qed
qed

lemma non_inf_int_gt: "non_inf {(a,b :: int) . a > b}" (is "non_inf ?r")
  by (rule non_inf_image[OF non_inf_rat_gt, of 1 _ of_int], auto simp: rat_gt_def)

lemma SN_imp_non_inf: "SN r \<Longrightarrow> non_inf r"
  by (intro non_infI, auto)


text {* for monotonicity, we do not use 0, 1, -1, and 2 as in CADE07, but a dedicated datatype *}

datatype dependance = Ignore | Increase | Decrease | Wild

type_synonym 'f dep = "('f \<times> nat) \<Rightarrow> nat \<Rightarrow> dependance"

fun rel_dep :: "'a rel \<Rightarrow> dependance \<Rightarrow> 'a rel" (infix "^^^" 52) where 
  "_ ^^^ Ignore = {}"
| "r ^^^ Increase = r"
| "r ^^^ Decrease = (r ^-1)"
| "r ^^^ Wild = (r^<->)"

lemma rel_dep_mono: "r1 \<subseteq> r2 \<Longrightarrow> r1 ^^^ k \<subseteq> r2 ^^^ k"
  by (cases k, auto)

fun invert_dep :: "dependance \<Rightarrow> dependance" where 
  "invert_dep Increase = Decrease"
| "invert_dep Decrease = Increase"  
| "invert_dep d = d"
  
lemma rel_dep_invert_mono: "r1^-1 \<subseteq> r2 \<Longrightarrow> r1 ^^^ invert_dep k \<subseteq> r2 ^^^ k"
  by (cases k, auto)

text {* for compatibility (monotonicity) we only consider F-terms, not all terms as in CADE07 *}

definition dep_compat :: "'f sig \<Rightarrow> ('f,'v)term rel \<Rightarrow> 'f dep \<Rightarrow> bool"
  where "dep_compat F r \<pi> \<equiv> \<forall> f bef s t aft. \<Union> funas_term ` ({s,t} \<union> set bef \<union> set aft) \<subseteq> F \<longrightarrow> 
  {(s,t)} ^^^ \<pi> (f,Suc (length bef + length aft)) (length bef) \<subseteq> r \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r"

lemma dep_compatE[elim]: assumes dep: "dep_compat F r \<pi>" shows "funas_args_term (Fun f (bef @ s # aft)) \<subseteq> F \<Longrightarrow> funas_term t \<subseteq> F \<Longrightarrow> 
  {(s,t)} ^^^ \<pi> (f,Suc (length bef + length aft)) (length bef) \<subseteq> r 
  \<Longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r"
  unfolding funas_args_term_def
  by (rule dep[unfolded dep_compat_def, rule_format, of s t bef aft f], force+)

lemma dep_compatI[intro]: assumes "\<And> f bef s t aft. 
  \<lbrakk>\<And> u. u \<in> {s,t} \<union> set bef \<union> set aft \<Longrightarrow> funas_term u \<subseteq> F\<rbrakk> \<Longrightarrow>
  {(s,t)} ^^^ \<pi> (f,Suc (length bef + length aft)) (length bef) \<subseteq> r 
  \<Longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r"
  shows "dep_compat F r \<pi>" 
  unfolding dep_compat_def
  by (intro allI impI, rule assms, auto)

text {* F-stability (closure under F-substitutions) is the same as in CADE07 *}  

definition F_subst_closed :: "'f sig \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
  where "F_subst_closed F r \<equiv> \<forall> \<sigma> s t. \<Union> funas_term ` range \<sigma> \<subseteq> F \<longrightarrow> (s,t) \<in> r \<longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> r"

lemma F_subst_closedI[intro]: assumes "\<And> \<sigma> s t. \<Union> funas_term ` range \<sigma> \<subseteq> F \<Longrightarrow> (s,t) \<in> r \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> r"
  shows "F_subst_closed F r" using assms unfolding F_subst_closed_def by blast

lemma F_subst_closedE[elim]: "F_subst_closed F r \<Longrightarrow> \<Union> funas_term ` range \<sigma> \<subseteq> F \<Longrightarrow> (s,t) \<in> r \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> r"
  unfolding F_subst_closed_def by auto

lemma F_subst_closed_UNIV: "F_subst_closed F r \<Longrightarrow> F = UNIV \<Longrightarrow> subst_closed r" unfolding F_subst_closed_def 
  using subst_closedI[of r] by auto

text {* a locale to have all nice properties together *}
locale non_inf_order =
  fixes S :: "('f,'v)trs"
  and NS :: "('f,'v)trs"
  and F :: "'f sig"
  and \<pi> :: "'f dep"
  assumes stable_NS: "F_subst_closed F NS"
  and stable_S: "F_subst_closed F S"
  and trans_NS: "trans NS"
  and refl_NS: "refl NS"
  and compat: "S O NS \<subseteq> S"
  and compat2: "NS O S \<subseteq> S"
  and trans_S: "trans S"
  and non_inf: "non_inf S"
  and dep_compat_NS: "dep_compat F NS \<pi>"

end

