(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory SubList
imports List2 Set Multiset2

begin

inductive sublist :: "'a list \<Rightarrow> 'a list \<Rightarrow> bool" where
  Equal: "sublist [] []"
| Skip : "\<And>as b bs. sublist as bs \<Longrightarrow> sublist as (b # bs)"
| Shift: "\<And>a as bs. sublist as bs \<Longrightarrow> sublist (a # as) (a # bs)"

lemma sublist_bottom[simp]: "sublist [] as"
proof (induct as)
case Nil thus ?case by (rule Equal)
next
case (Cons a as) thus ?case by (rule Skip)
qed

lemma sublist_bottom_unique:
  assumes "sublist as []"
  shows "as = []"
using assms by (cases rule: sublist.cases) auto

lemma length_sublist:
  assumes "sublist as bs"
  shows "length as \<le> length bs"
using assms by (induct rule: sublist.induct) auto

lemma sublist_refl[simp]: "sublist as as"
proof (induct as)
case Nil thus ?case by (rule Equal)
next
case (Cons a as) thus ?case by (rule Shift)
qed

lemma prefix_sublist: "sublist as (as @ bs)"
proof (induct as)
case Nil thus ?case by simp
next
case (Cons a as) thus ?case by (simp, rule Shift)
qed

lemma suffix_sublist: "sublist bs (as @ bs)"
proof (induct as)
case Nil thus ?case by simp
next
case (Cons a as) thus ?case by (simp, rule Skip)
qed

lemma remove1_sublist: "sublist (remove1 x as) as"
proof (induct as)
case Nil thus ?case by simp
next
case (Cons a as)
  show ?case 
  proof (cases "x = a") 
  case True with Skip[of as as] show ?thesis by simp
  next
  case False with Shift[OF Cons] show ?thesis by simp
  qed
qed

lemma filter_sublist: "sublist [y\<leftarrow> as. f y] as"
proof (induct as)
case Nil thus ?case by simp
next
case (Cons a as)
  show ?case 
  proof (cases "f a") 
  case True with Shift[OF Cons] show ?thesis by simp
  next
  case False with Skip[OF Cons] show ?thesis by simp
  qed
qed

lemma sublist_trans:
  assumes "sublist as bs"
  and "sublist bs cs"
  shows "sublist as cs"
using assms(2) assms(1) proof (induct arbitrary: as rule: sublist.induct)
case (Equal as) thus ?case by assumption
next
case (Skip bs c cs as) from Skip(2)[OF Skip(3)] show ?case by (rule sublist.Skip)
next
case (Shift b bs cs as) note IH = this
  show ?case using IH(3)
  proof (cases rule: sublist.cases)
  case Skip from IH(2)[OF Skip] show ?thesis by (rule sublist.Skip)
  next
  case (Shift as) from IH(2)[OF Shift(2)] Shift(1) show ?thesis by (simp add: sublist.Shift)
  qed
qed

lemma sublist_append:
  assumes "sublist as cs"
  and "sublist bs ds"
  shows "sublist (as @ bs) (cs @ ds)"
using assms(1) proof (induct rule: sublist.induct)
case Equal from assms(2) show ?case by simp
next
case (Skip as c cs)
  from sublist.Skip[OF Skip(2)] show ?case by simp
next
case (Shift a as cs)
  from sublist.Shift[OF Shift(2)] show ?case by simp
qed

lemma sublist_subset:
  assumes "sublist as bs"
  shows "set as \<le> set bs"
using assms proof (induct rule: sublist.induct)
case Equal show ?case by simp
next
case (Skip as b bs) from subset_trans[OF Skip(2)] show ?case by (cases "b \<in> set bs") auto
next
case (Shift a as bs) thus ?case by (cases "a \<in> set as") auto
qed

lemma sublist_submultiset:
  assumes "sublist as bs"
  shows "multiset_of as \<le> multiset_of bs"
using assms proof (induct rule: sublist.induct)
case Equal show ?case by simp
next
case (Skip as b bs) from mset_le_trans[OF Skip(2)] show ?case by simp
next
case (Shift a as bs) thus ?case by simp
qed

lemma left_extension_not_sublist:
  assumes "as \<noteq> []"
  shows "\<not> sublist (as @ bs) bs"
proof-
{
  assume Hf: "sublist (as @ bs) bs"
  with length_sublist[OF this] length_append assms
   have False by simp
}
thus ?thesis by fast
qed

lemma right_extension_not_sublist:
  assumes "as \<noteq> []"
  shows "\<not> sublist (bs @ as) bs"
proof-
{
  assume Hf: "sublist (bs @ as) bs"
  with length_sublist[OF this] length_append assms
   have False by simp
}
thus ?thesis by fast
qed

lemma sublist_antisym:
  assumes "sublist as bs"
  and "sublist bs as"
  shows "as = bs"
using assms proof (induct rule: sublist.induct)
case Equal thus ?case by simp
next
case (Skip as b bs) note IH = this
  show ?case using IH(3)
  proof (cases rule: sublist.cases)
  case (Skip a as)
    from sublist_trans[OF IH(1) sublist_trans[OF SubList.sublist.Skip[OF sublist_refl[of bs], of b]
     Skip(2)]] Skip(1) left_extension_not_sublist[of "[ a ]" as] show ?thesis by simp
  next
  case (Shift as)
    from sublist_trans[OF IH(1) Shift(2)] Shift(1) left_extension_not_sublist[of "[ b ]" as]
     show ?thesis by simp
  qed
next
case (Shift a as bs) note IH = this
  show ?case using IH(3)
  proof (cases rule: sublist.cases)
  case Skip
    from sublist_trans[OF Skip IH(1)] left_extension_not_sublist[of "[ a ]" bs]
     show ?thesis by simp
  next
  case Shift
    from IH(2)[OF Shift] show ?thesis by simp
  qed
qed

end