(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Substitution
imports
  Term_Impl
  Option_Monad
begin

section {* Unification *}

type_synonym ('f, 'v) equation = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) equations = "('f, 'v) equation set"

definition unifiers :: "('f, 'v) equations \<Rightarrow> ('f, 'v) subst set" where
  "unifiers eqs \<equiv> {\<sigma>. \<forall>p \<in> eqs. fst p \<cdot> \<sigma> = snd p \<cdot> \<sigma>}"

definition is_mgu :: "('f, 'v) subst \<Rightarrow> ('f, 'v) equations \<Rightarrow> bool" where
  "is_mgu \<sigma> eqs \<equiv> (\<sigma> \<in> unifiers eqs \<and> (\<forall>\<tau> \<in> unifiers eqs. (\<exists>\<gamma> . \<tau> = \<sigma> \<circ>s \<gamma>)))"

definition is_smgu :: "('f, 'v) subst \<Rightarrow> ('f, 'v) equations \<Rightarrow> bool" where
  "is_smgu \<sigma> eqs \<equiv> (\<sigma> \<in> unifiers eqs \<and> (\<forall>(\<tau>::('f, 'v) subst) \<in> unifiers eqs. \<tau> = \<sigma> \<circ>s \<tau>))"

lemma is_smgu_imp_is_mgu:
  "is_smgu \<sigma> eqs \<Longrightarrow> is_mgu \<sigma> eqs"
  by (auto simp: is_mgu_def is_smgu_def)

lemma unifiers_Un[simp]: "unifiers (s \<union> t) = unifiers s \<inter> unifiers t"
  by (auto simp: unifiers_def)

lemma unifiers_insert: (* simp not added for readability *)
  "unifiers (insert p t) = {\<sigma>. fst p \<cdot> \<sigma> = snd p \<cdot> \<sigma>} \<inter> unifiers t"
  by (auto simp: unifiers_def)

lemma unifiers_swap:
  "unifiers (insert (s, t) eqs) = unifiers (insert (t, s) eqs)"
  by (auto simp: unifiers_insert)

lemma unifiers_orient[simp]:
  "unifiers (insert (Fun f ss, Var x) eqs) = unifiers (insert (Var x, Fun f ss) eqs)"
  by (simp add: unifiers_swap)

lemma unifiers_empty[simp]: "unifiers {} = UNIV"
  by (auto simp: unifiers_def)

lemma unifiers_insert_ident[simp]: "unifiers (insert (t, t) eqs) = unifiers eqs" 
  by (auto simp: unifiers_insert)

lemma unifiers_Int1[simp]:
  "(s, t) \<in> eqs \<Longrightarrow> unifiers {(s, t)} \<inter> unifiers eqs = unifiers eqs"
  by (auto simp: unifiers_def)

lemma unifiers_unify: "\<sigma> \<in> unifiers eqs \<Longrightarrow> eq \<in> eqs \<Longrightarrow> fst eq \<cdot> \<sigma> = snd eq \<cdot> \<sigma>"
  by (auto simp: unifiers_def)

lemma unifiers_occur[simp]: fixes t :: "('f,'v)term"
  assumes tx: "t \<noteq> Var x" and xt: "x \<in> vars_term t"
  shows "unifiers (insert (Var x, t) eqs) = {}"
using assms
proof -
  from xt have "t \<unrhd> Var x" by auto
  hence "t \<rhd> Var x \<or> t = Var x" by auto
  with tx have subt: "t \<rhd> Var x" by simp
  show ?thesis 
  proof (unfold unifiers_def, rule set_eqI, auto)
    fix \<sigma> :: "('f,'v)subst"
    assume eq: "\<sigma> x = t \<cdot> \<sigma>"
    from subt have "t \<cdot> \<sigma> \<rhd> Var x \<cdot> \<sigma>" by (rule supt_stable)
    hence "t \<cdot> \<sigma> \<noteq> Var x \<cdot> \<sigma>" by auto
    with eq have False by simp
    thus "\<exists>p\<in>eqs. fst p \<cdot> \<sigma> \<noteq> snd p \<cdot> \<sigma>" ..
  qed
qed


subsection {* More Properties of Substitutions *}

lemma subst_subst_domain[simp]: "subst_domain (subst x t) = (if (t = Var x) then {} else {x})"
proof -
  have "subst_domain (subst x t) = (if t = Var x then {} else {x})" 
  proof (rule set_eqI, simp only: subst_domain_def)
    fix y 
    show "(y \<in> {y. subst x t y \<noteq> Var y}) = (y \<in> (if t = Var x then {} else {x}))"
     by (cases "x = y", auto simp: subst_def)
 qed
  thus ?thesis .
qed

lemma subst_subst_range[simp]: "subst_range (subst x t) = (if (t = Var x) then {} else {t})"
  by (cases "t = Var x", auto simp: subst_domain_def subst_def)

lemma subst_ident[simp]: "s \<cdot> subst x (Var x) = s" (is "_ \<cdot> ?xx = _")
proof -
  have "?xx = Var" by (simp add: subst_def fun_eq_iff)
  thus ?thesis by simp
qed

lemma subst_ident2: assumes x: "x \<notin> vars_term t" shows "t \<cdot> subst x u = t"
proof -
  have "t \<cdot> subst x u = t \<cdot> Var"
    by (rule term_subst_eq, insert x, auto simp: subst_def)
  thus ?thesis by simp
qed

lemma subst_vars: "vars_term (t \<cdot> \<sigma>) \<subseteq> vars_term t - subst_domain \<sigma> \<union> \<Union> vars_term`subst_range \<sigma> "
proof -
  {
    fix x y 
    assume x: "x \<in> vars_term (\<sigma> y)"
    and all: "\<forall>a. \<sigma> a = Var a \<or> x \<notin> vars_term (\<sigma> a)"
    and neq: "\<sigma> x \<noteq> Var x"
    from spec[OF all, of y] neq x 
    have False by auto
  } 
  thus ?thesis
  unfolding vars_term_subst 
  by (force simp: subst_domain_def)
qed

lemma subst_vars2:
  assumes elem: "x \<in> vars_term(t \<cdot> \<sigma>)" shows "x \<in> vars_term t - subst_domain \<sigma> \<union> \<Union> vars_term`subst_range \<sigma> "
proof -
  have "vars_term(t \<cdot> \<sigma>) \<subseteq> vars_term t - subst_domain \<sigma> \<union> \<Union> vars_term`subst_range \<sigma> " by (rule subst_vars)
  with elem show ?thesis by auto
qed
 
lemma subst_vanish: assumes xt: "x \<notin> vars_term t" shows "x \<notin> vars_term(u \<cdot> subst x t)"
proof 
  let ?xt = "subst x t"
  assume xv: "x \<in> vars_term(u \<cdot> ?xt)"
  show False
  proof (cases "t = Var x")
    case True with xt show False by simp
  next
    case False
    from xv have "x \<in> vars_term u - subst_domain ?xt \<union> \<Union> vars_term`subst_range ?xt" by (rule subst_vars2)
    with False and xt show False by (simp add: subst_domain_def subst_def)
  qed
qed

lemma subst_invert1:
  assumes yt: "y \<notin> vars_term t" and ysxt: "y \<in> vars_term(s \<cdot> subst x t)" shows "y \<in> vars_term s"
using assms 
proof (cases "t = Var x", auto) 
  assume tx: "t \<noteq> Var x" 
  from ysxt show "y \<in> vars_term s"
  proof (induct s)
    case (Var z) with tx and yt show ?case by (cases "z = x", auto simp: subst_def)
  next
    case (Fun f ss) with yt show ?case by (auto simp: subst_def)
  qed
qed

lemma subst_invert2: assumes "y \<notin> vars_term t" and "x \<noteq> y" and "s \<cdot> subst x t \<noteq> Var y" shows
  "s \<noteq> Var y" 
using assms
by (cases s, auto simp: subst_def)

lemma subst_invert3: assumes "y \<notin> vars_term t" and "s \<cdot> subst x t = Var y" shows "s = Var y" 
proof (cases s)
  case (Fun f ss)
  thus ?thesis using assms by (auto simp: subst_def)
next
  case (Var z)
  thus ?thesis using assms by (cases "x=z", auto simp: subst_def)
qed

lemma subst_Fun[simp]:
  "(Fun f ss \<cdot> \<sigma> = Fun g ts \<cdot> \<sigma>) =
    (f = g \<and> length ss = length ts \<and> \<sigma> \<in> unifiers (set (zip ss ts)))" 
using assms
proof -
  have "(Fun f ss \<cdot> (\<sigma>) = Fun g ts \<cdot> (\<sigma>)) =
    (f = g \<and> length ss = length ts \<and> (\<sigma>) \<in> unifiers (set (zip ss ts)))" 
  proof (induct ss arbitrary: ts)
    case Nil thus ?case by simp
  next
    case (Cons s sss)
    thus ?case by (cases ts, auto simp: unifiers_insert)
  qed
  thus ?thesis by simp
qed

section {* Computing Most General Unifiers *}

(* x occurs somewhere within a list of term-eqs *)
definition varsL :: "('f, 'v) equation list \<Rightarrow> 'v \<Rightarrow> bool" where
  "varsL list x \<equiv> foldr (\<lambda>eq b.
    x \<in> set (vars_term_list (fst eq)) \<or> x \<in> set (vars_term_list (snd eq)) \<or> b) list False"

(* substitution {x / t} applied on term-pair list *)
definition
  simple_subst_list :: "'v \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) equation list \<Rightarrow> ('f, 'v) equation list"
where
  "simple_subst_list x t list =
    map (\<lambda>p. (fst p \<cdot> subst x t,snd p \<cdot> subst x t)) list"

definition
  decompose :: "('f,'v) term \<Rightarrow> ('f,'v) term \<Rightarrow> ('f, 'v) equation list option"
where
  "decompose s t =
    (case (s, t) of
      (Fun f ss, Fun g ts) \<Rightarrow> if f = g then zip_option ss ts else None
    | _ \<Rightarrow> None)"

(* unification of list of equations of terms *)
function
  unify :: "('f, 'v) equation list \<Rightarrow> ('f, 'v) equation list \<Rightarrow> ('f, 'v) equation list option"
where
  "unify [] seen = Some seen"
| "unify ((Fun f ss, Fun g ts) # unifs) seen =
    (case decompose (Fun f ss) (Fun g ts) of
      None \<Rightarrow> None
    | Some eqs \<Rightarrow> unify (eqs @ unifs) seen)"
| "unify ((Fun f ss, Var x) # unifs) seen = unify ((Var x, Fun f ss) # unifs) seen"
| "unify ((Var x , t) # unifs) seen = (if t = Var x then unify unifs seen
    else if x \<in> set (vars_term_list t) then None 
    else unify (simple_subst_list x t unifs) ((Var x, t) # simple_subst_list x t seen))"
  by pat_completeness auto

(* extracting a substitution from a list of equations *)
fun subst_of :: "('f, 'v) equation list \<Rightarrow> ('f, 'v) subst" where
  "subst_of [] = Var"
| "subst_of ((Var x,t) # ts) = (\<lambda>y. if (x = y)
    then t
    else subst_of ts y)"

(* computing the mgu of two terms *)
fun mgu :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) subst option" where
  "mgu s t =
    (case unify [(s,t)] [] of   
      None \<Rightarrow> None
    | Some res \<Rightarrow> Some (subst_of res))"

(* the mgu of two terms (to be used if it is guaranteed that the terms are unifiable) *)
definition the_mgu :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f ,'v) subst" where
  "the_mgu s t \<equiv> (case mgu s t of None \<Rightarrow> Var | Some \<delta> \<Rightarrow> \<delta>)"

lemma simple_subst_list_set[simp]:
  "set (simple_subst_list x t eqs) = (\<lambda>p. (fst p \<cdot> subst x t, snd p \<cdot> subst x t)) ` (set eqs)"
  by (simp add: simple_subst_list_def)

definition vars_eqs where
  "vars_eqs zs \<equiv> (\<Union> p \<in> zs . vars_term(fst p) \<union> vars_term(snd p))"

lemma varsL_sound[simp]: "varsL l x = (x \<in> vars_eqs (set l))"
proof -
  have "foldr (\<lambda> pair b. x \<in> set (vars_term_list (fst pair)) \<or> x \<in> set (vars_term_list (snd pair)) \<or> b) l False = (x \<in> vars_eqs (set l))"
  proof (induct l)
    case Nil thus ?case by (simp add: vars_eqs_def)
  next
    case (Cons pair ll)
    thus ?case 
    proof (cases pair)
      case (Pair s t)
      with Cons vars_term_list_sound show ?thesis 
	by (cases "x \<in> set (vars_term_list s) \<or> x \<in> set (vars_term_list t)", auto simp: vars_eqs_def)
    qed
  qed
  thus ?thesis by (simp add: varsL_def)
qed

lemma decompose_args: "decompose (Fun f ss) (Fun g ts) = Some eqs \<Longrightarrow> eqs = zip ss ts" 
  by (cases "f=g", auto simp: decompose_def)

lemma decompose_roots: "decompose (Fun f ss) (Fun g ts) = Some eqs \<Longrightarrow> f = g" 
  by (cases "f=g", auto simp: decompose_def)

lemma decompose_length: "decompose (Fun f ss) (Fun g ts) = Some eqs \<Longrightarrow> length ss = length ts"
  by (cases "f=g", auto simp: decompose_def)

lemma decompose_None:
  "decompose (Fun f ss) (Fun g ts) = None \<Longrightarrow> f \<noteq> g \<or> length ss \<noteq> length ts"
  by (cases "f=g", auto simp: decompose_def)

lemma decompose_sound1[simp]:
  assumes dec: "decompose (Fun f ss) (Fun g ts) = None"
  shows "unifiers (insert((Fun f ss, Fun g ts)) eqs) = {}"
proof - 
  from dec have "f \<noteq> g \<or> length ss \<noteq> length ts" by (rule decompose_None) 
  thus ?thesis by (auto simp:unifiers_insert List.map_eq_imp_length_eq[where ?xs = "ss" and ?ys = "ts"])
qed

lemma decompose_sound2[simp]:
  assumes dec: "decompose (Fun f ss) (Fun g ts) = Some eqs" 
  shows "unifiers (insert ((Fun f ss, Fun g ts)) rest) = unifiers (set eqs \<union> rest)"
using assms
proof - 
  from dec have fg: "f = g" by (rule decompose_roots) 
  from dec have len: "length ss = length ts" by (rule decompose_length) 
  from dec have zip: "eqs = zip ss ts" by (rule decompose_args)
  from len have "{\<sigma>. map (\<lambda>t. t \<cdot> \<sigma>) ss = map (\<lambda>t. t \<cdot> \<sigma>) ts} = unifiers (set (zip ss ts))" 
  proof (induct ss arbitrary: ts)
    case Nil thus ?case by (cases ts, auto simp: unifiers_def)
  next
    case (Cons s ss) note oCons = this
    thus ?case 
    proof (cases ts)
      case Nil with Cons have False by simp thus ?thesis ..
    next
      case (Cons t ts) 
      with oCons show ?thesis by (auto simp: unifiers_def)
    qed
  qed
  with fg zip show ?thesis by (auto simp: unifiers_insert) 
qed

(* towards termination of unify *)

(* mesures for termination *)
definition
  m1 :: "((('f,'v) term \<times> ('f,'v) term) list) \<times> ((('f,'v) term \<times> ('f,'v) term) list) \<Rightarrow> nat"
where
  "m1 \<equiv> \<lambda> (l1,l2). card (vars_eqs (set(l1)))"

definition
  m2 :: "((('f,'v) term \<times> ('f,'v) term) list) \<times> ((('f,'v) term \<times> ('f,'v) term) list) \<Rightarrow> nat"
where
  "m2 \<equiv> \<lambda> (l1,l2). listsum (map (\<lambda> (s,t). Suc (size s + size t) ) l1)"

definition
  m3 :: "((('f,'v) term \<times> ('f,'v) term) list) \<times> ((('f,'v) term \<times> ('f,'v) term) list) \<Rightarrow> nat"
where
  "m3 \<equiv> \<lambda> (l1,l2). length (filter (\<lambda> (s,t). is_Fun s \<and> is_Var t) l1)"

(* termination can be shown by lex. combination of (m1,m2,m3) *)
  

lemma m1subset:
  assumes subset: "vars_eqs (set (r1)) \<subseteq> vars_eqs (set (l1))"
  shows "m1 (r1,r2) \<le> m1 (l1,l2)" unfolding m1_def split
proof (rule card_mono[OF _ subset])
  show "finite (vars_eqs (set l1))" unfolding vars_eqs_def by auto 
qed

lemma m1psubset:
  assumes subset: "vars_eqs (set (r1)) \<subset> vars_eqs (set l1)"
  shows "m1 (r1,r2) < m1 (l1,l2)" unfolding m1_def split
proof (rule psubset_card_mono[OF _ subset])
  show "finite (vars_eqs (set l1))" unfolding vars_eqs_def by auto 
qed
  
lemma vars_eqs_extract: "x \<in> vars_eqs eqs \<Longrightarrow> \<exists> p \<in> eqs. x \<in> vars_term (fst p) \<or> x \<in> vars_term (snd p)"
  by (unfold vars_eqs_def, auto) 

lemma vars_eqs_Un[simp]: "vars_eqs (ps \<union> qs) = vars_eqs ps \<union> vars_eqs qs" 
  by (unfold vars_eqs_def, auto)

lemma vars_eqs_insert[simp]:
  "vars_eqs (insert pair ps) = vars_term(fst pair) \<union> vars_term(snd pair) \<union> vars_eqs ps"
  by (unfold vars_eqs_def, auto)

lemma vars_eqs_empty[simp]: "vars_eqs {} = {}"
  by (unfold vars_eqs_def, auto)

lemma vars_eqs_Fun[simp]: "x \<in> vars_eqs (set (zip ss ts)) \<Longrightarrow> x \<in> vars_term(Fun f ss) \<or> x \<in> vars_term(Fun g ts)" 
proof (unfold vars_eqs_def, induct ss arbitrary: ts)
  case Nil thus ?case by simp
next
  case (Cons s ss) note oCons = this
  thus ?case
  proof (cases ts)
    case Nil with Cons show ?thesis by simp
  next
    case (Cons t tts)
    with oCons show ?thesis by auto
  qed
qed

termination unify
proof -
  let ?R = "measures [m1,m2,m3]"
  {
    fix l1 l2 r1 r2 :: "(('f,'v) term \<times> ('f,'v) term) list"
    assume uns: "vars_eqs (set r1) \<subseteq> vars_eqs (set l1)" 
      and mea: "((r1,r2),l1,l2) \<in> measures [m2,m3]"
    have "((r1,r2),l1,l2) \<in> ?R" 
    proof (cases "vars_eqs (set r1) \<subset> vars_eqs (set l1)")
      case True
      hence "m1 (r1,r2) < m1 (l1,l2)" by (rule m1psubset)
      thus ?thesis by (rule measures_less)
    next
      case False 
      from uns have "m1 (r1,r2) \<le> m1 (l1,l2)" by (rule m1subset)
      thus ?thesis 
        by (rule measures_lesseq, insert mea, blast) 
    qed
  } note m1step = this
  show ?thesis 
  proof  
    show "wf ?R" by simp  
  next
  (* case 3 *)
    fix x t and unifs seen :: "(('f,'v) term \<times> ('f,'v) term) list"
    show "((unifs, seen), (Var x, t) # unifs, seen) \<in> ?R"
      by (rule m1step[OF _ measures_less], auto simp: m2_def) 
  next
  (* case 2 *)
    fix f ss x and unifs seen :: "(('f,'v) term \<times> ('f,'v) term) list"
    show "(((Var x, Fun f ss) # unifs, seen), (Fun f ss, Var x) # unifs, seen) \<in> ?R"
      by (rule m1step[OF _ measures_lesseq], auto simp: m2_def m3_def)
  next
  (* case 1 *)    
    fix f ss g ts and unifs seen eqs :: "(('f,'v) term \<times> ('f,'v) term) list"
    have main: "((zip ss ts @ unifs, seen), (Fun f ss, Fun g ts) # unifs, seen) \<in> ?R"
    proof (rule m1step[OF _ measures_less])
      show "m2 (zip ss ts @ unifs, seen) < m2 ((Fun f ss, Fun g ts) # unifs, seen)"
        unfolding m2_def split
      proof (induct ss arbitrary: ts)
        case (Cons b sss) 
        from this[of "tl ts"]
        show ?case by (cases ts, auto)
      qed simp      
    qed (auto simp: set_zip vars_eqs_def)
    assume "decompose (Fun f ss) (Fun g ts) = Some eqs"
    hence "eqs = zip ss ts" by (rule decompose_args)
    with main show "((eqs @ unifs, seen), (Fun f ss, Fun g ts) # unifs, seen) \<in> ?R" by simp    
  next
  (* case 4 *)
    fix x :: 'v and  t  :: "('f, 'v) term" and unifs :: "(('f,'v) term \<times> ('f,'v) term) list" and seen :: "(('f,'v) term \<times> ('f,'v) term) list"
    assume "\<not> x \<in> set (vars_term_list t)" 
    hence x: "x \<notin> vars_term t" by simp
    let ?xt = "\<lambda> u. u \<cdot> subst x t"
    let ?xtp = "\<lambda> p. (?xt (fst p), ?xt (snd p))"
    show "((simple_subst_list x t unifs, (Var x, t) # simple_subst_list x t seen), (Var x, t) # unifs, seen) \<in> ?R"
    proof (rule measures_less, rule m1psubset, unfold simple_subst_list_set)
      show "vars_eqs (?xtp ` set unifs) \<subset> vars_eqs (set ((Var x, t) # unifs))" (is "?L \<subset> ?R")
      proof
        have "x \<in> ?R" by auto
        moreover have "x \<notin> ?L"
        proof
          assume "x \<in> ?L"
          then obtain u where "x \<in> vars_term (?xt u)" by (auto simp: vars_eqs_def)
          from this[unfolded vars_term_subst] x show False by (cases "x \<in> vars_term u", auto simp: subst_def)
        qed
        ultimately show "?L \<noteq> ?R" by auto
      next
        show "?L \<subseteq> ?R"
        proof
          fix y
          assume "y \<in> ?L"
          then obtain u v where uv: "(u,v) \<in> set unifs \<or> (v,u) \<in> set unifs" and y: "y \<in> vars_term (?xt u)" 
            unfolding vars_eqs_def by force
          from y[unfolded vars_term_subst] have "y \<in> vars_term u \<or> y \<in> vars_term t" 
            by (cases "x \<in> vars_term u", auto simp: subst_def)
          thus "y \<in> ?R" using uv unfolding vars_eqs_def by force
        qed
      qed
    qed
  qed
qed

hide_const m1 m2 m3
	


(* before proving soundness, first have some better induction scheme *) 

(* we first need an ugly one with \<forall> due to the types ! *)

lemma fixes P :: "(('f,'v) term \<times> ('f,'v) term) list \<Rightarrow> (('f,'v) term \<times> ('f,'v) term) list \<Rightarrow> (('f,'v) term \<times> ('f,'v) term) list option \<Rightarrow> bool"
  and res :: "(('f,'v) term \<times> ('f,'v) term) list option" 
  and xs :: "(('f,'v) term \<times> ('f,'v) term) list"
  and ys :: "(('f,'v) term \<times> ('f,'v) term) list"
assumes  
  "(\<forall> ys . res = Some ys \<longrightarrow> P [] ys res)    \<and>
  (\<forall> f ss g ts zs ys . (P (zip ss ts @ zs) ys res \<and> decompose (Fun f ss) (Fun g ts) = Some (zip ss ts) \<and> f = g \<and> length ss = length ts) \<longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res)  \<and>
  (\<forall> f ss g ts zs ys . res = None \<and> decompose (Fun f ss) (Fun g ts) = None  \<longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res)    \<and>
  (\<forall> f ss x zs ys . (P ((Var x, Fun f ss) # zs) ys res) \<longrightarrow> P ((Fun f ss, Var x) # zs) ys res)    \<and>
  (\<forall> x zs ys . P zs ys res \<longrightarrow> P ((Var x, Var x) # zs) ys res)    \<and>
  (\<forall> x t zs ys .  (res = None \<and> Var x \<noteq> t \<and> x \<in> vars_term t) \<longrightarrow> P ((Var x, t) # zs) ys res)    \<and>
  (\<forall> x t zs ys .  (P (map (\<lambda> p. (fst p \<cdot> subst x t, snd p \<cdot> subst x t)) zs) ((Var x, t) # map (\<lambda> p. (fst p \<cdot> subst x t, snd p \<cdot> subst x t)) ys) res \<and> t \<noteq> Var x \<and> x \<notin> vars_term t) \<longrightarrow> P ((Var x, t) # zs) ys res)"
  shows unify_pre_induct[rule_format]: "unify xs ys = res \<longrightarrow> P xs ys res"
proof (induct xs ys rule: unify.induct)
  case 1 thus ?case using assms by auto
next
  case (2 f ss g ts unifs simp) 
  thus ?case 
  proof (cases "decompose (Fun f ss) (Fun g ts)")
    case None with 2 assms show ?thesis by auto
  next
    case (Some eqs)
    with 2 have "unify (eqs @ unifs) simp = res \<longrightarrow> P (eqs @ unifs) simp res" by simp
    with 2 have a: "unify (eqs @ unifs) simp = res \<longrightarrow> P (eqs @ unifs) simp res" by simp
    from 2 assms have b: "(\<forall>f ss g ts zs ys . P (zip ss ts @ zs) ys res \<and> decompose (Fun f ss) (Fun g ts) = Some (zip ss ts) \<and> f = g \<and> length ss = length ts \<longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res)" by simp
    from Some have pz: "eqs = zip ss ts" by (rule decompose_args)
    from Some have len: "length ss = length ts" by (rule decompose_length)
    from Some have fg: "f = g" by (rule decompose_roots)
    from Some a b pz len fg show ?thesis by simp
  qed
next
  case (3 f ss x unifs seen) thus ?case using assms by simp
next
  case (4 x t unifs seen)
  let ?xt = "subst x t"
  show ?case
  proof (cases "t = Var x")
    case True
    with 4 have imp: "unify unifs seen = res \<longrightarrow> P unifs seen res" by auto
    from 4 assms have one: "(\<forall>x zs ys. P zs ys res \<longrightarrow> P ((Var x, Var x) # zs) ys res)" by simp
    from True one have "P unifs seen res \<longrightarrow> P ((Var x, t) # unifs) seen res" by auto
    with imp True show ?thesis by auto
  next
    case False note oFalse = this
    hence nElem: "t \<noteq> Var x" .
    show ?thesis 
    proof (cases "x \<in> set (vars_term_list t)")
      case True
      hence xvt: "x \<in>  vars_term t" using vars_term_list_sound by (auto)
      from  4 assms have "(\<forall>x t zs ys . res = None \<and> Var x \<noteq> t \<and> x \<in> vars_term t \<longrightarrow> P ((Var x, t) # zs) ys res)" by simp
      with xvt and nElem True show ?thesis by auto
    next
      case False
      hence xnt: "x \<notin> vars_term t \<and> \<not> x \<in> set (vars_term_list t)" using vars_term_list_sound by (auto)
    	with 4 nElem assms have imp: "P (map (\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) unifs) ((Var x, t) # 
        map (\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) seen) res  \<Longrightarrow>
      P ((Var x, t) # unifs) seen res" by simp
      from 4 nElem xnt have imp2: "unify (simple_subst_list x t unifs) ((Var x, t) # simple_subst_list x t seen) = res 
        \<Longrightarrow> P (simple_subst_list x t unifs) ((Var x, t) # simple_subst_list x t seen) res" by simp
      show ?thesis
        by (intro impI, rule imp, insert imp2 xnt, auto simp: simple_subst_list_def)
    qed
  qed 
qed 
	  
(* now here is the induction scheme you should use *)

lemma  
  assumes Done: "\<And>ys. res = Some ys \<Longrightarrow> P [] ys res" 
  and FF: "\<And>f ss g ts zs ys. \<lbrakk>P (zip ss ts @ zs) ys res; decompose (Fun f ss) (Fun g ts) = Some (zip ss ts); f = g; length ss = length ts\<rbrakk> \<Longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res"
  and Clash: "\<And>f ss g ts zs ys. \<lbrakk>res = None; decompose (Fun f ss) (Fun g ts) = None\<rbrakk>  \<Longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res"
  and Swap: "\<And>f ss x zs ys. P ((Var x, Fun f ss) # zs) ys res \<Longrightarrow> P ((Fun f ss, Var x) # zs) ys res"
  and XX: "\<And>x zs ys. P zs ys res \<Longrightarrow> P ((Var x, Var x) # zs) ys res"
  and Occur: "\<And>x t zs ys.  \<lbrakk>res = None; Var x \<noteq> t; x \<in> vars_term t\<rbrakk> \<Longrightarrow> P ((Var x, t) # zs) ys res"
  and Subst: "\<And>x t zs ys.  \<lbrakk>P (map (\<lambda>p. (fst p \<cdot> subst x t, snd p \<cdot> subst x t)) zs) ((Var x, t) # map (\<lambda>p. (fst p \<cdot> subst x t, snd p \<cdot> subst x t)) ys) res; t \<noteq> Var x; x \<notin> vars_term t\<rbrakk> \<Longrightarrow> P ((Var x, t) # zs) ys res"
  shows unify_induct [case_names Done FF Clash Swap XX Occur Subst]: "unify xs ys = res \<longrightarrow> P xs ys res"
proof
  assume ass: "unify xs ys = res" 
  have "(\<forall>ys. res = Some ys \<longrightarrow> P [] ys res) \<and>
 (\<forall>f ss g ts zs ys.
     P (zip ss ts @ zs) ys res \<and> decompose (Fun f ss) (Fun g ts) = Some (zip ss ts) \<and> f = g \<and> length ss = length ts \<longrightarrow>
     P ((Fun f ss, Fun g ts) # zs) ys res) \<and>
 (\<forall>f ss g ts zs ys. res = None \<and> decompose (Fun f ss) (Fun g ts) = None \<longrightarrow> P ((Fun f ss, Fun g ts) # zs) ys res) \<and>
 (\<forall>f ss x zs ys. P ((Var x, Fun f ss) # zs) ys res \<longrightarrow> P ((Fun f ss, Var x) # zs) ys res) \<and>
 (\<forall>x zs ys. P zs ys res \<longrightarrow> P ((Var x, Var x) # zs) ys res) \<and>
 (\<forall>x t zs ys. res = None \<and> Var x \<noteq> t \<and> x \<in> vars_term t \<longrightarrow> P ((Var x, t) # zs) ys res) \<and>
 (\<forall>x t zs ys.
     P (map (\<lambda>p. (fst p \<cdot> subst x t, snd p \<cdot> subst x t)) zs) ((Var x, t) # map (\<lambda>p. (fst p \<cdot> subst x t, snd p \<cdot> subst x t)) ys) res \<and>
     t \<noteq> Var x \<and> x \<notin> vars_term t \<longrightarrow> P ((Var x, t) # zs) ys res)" using assms by auto
  from this and ass show "P xs ys res" by (rule unify_pre_induct)
qed

definition solved :: "('f, 'v) equations \<Rightarrow> bool" where
  "solved eqs \<equiv> \<forall>(s,t) \<in> eqs. \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> vars_eqs (eqs - {(s,t)})"

lemma solved_mono: assumes sub: "xs \<subseteq> ys" and solve: "solved ys" shows "solved xs"
proof (unfold solved_def, rule ballI)
  fix p
  assume "p \<in> xs"
  show "(\<lambda>(s, t). \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> vars_eqs (xs - {(s, t)})) p"
  proof (cases p)
    case (Pair s t)
    with `p \<in> xs` and sub have "(s,t) \<in> ys" by force
    with solve have "\<exists> x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> vars_eqs (ys - {(s,t)})" by (unfold solved_def, auto)
    from this obtain x where "s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> vars_eqs (ys - {(s,t)})" ..
    with sub have "s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> vars_eqs (xs - {(s,t)})" by (unfold vars_eqs_def, auto)
    with Pair show ?thesis by force 
  qed
qed

lemma solvedMainProperty: fixes eqs :: "(('f,'v) term \<times> ('f,'v) term) list"
  assumes solved: "solved (set eqs)"
  shows "subst_of eqs \<in> unifiers (set eqs) \<and> (\<forall>\<sigma> \<in> unifiers (set eqs). \<sigma> = subst_of eqs \<circ>s \<sigma>)"
proof -
  let ?S = "subst_of eqs"  
  have "solved (set eqs \<union> set eqs) \<longrightarrow> (\<exists> (f :: 'v \<Rightarrow> ('f,'v) term). ?S = f \<and> (\<forall> p \<in> set eqs. f (the_Var (fst p)) = snd p) \<and> 
    (\<forall> \<sigma> \<in> unifiers (set eqs). \<sigma> = ?S \<circ>s \<sigma>) \<and> (subst_domain ?S \<inter> \<Union> ((vars_term \<circ> snd) ` set eqs) \<subseteq> {}))"
  proof (induct eqs rule: list.induct[where ?P = "\<lambda> list. solved (set eqs \<union> set list) \<longrightarrow> 
      (\<exists> (f :: 'v \<Rightarrow> ('f,'v) term). subst_of list = f \<and> (\<forall> p \<in> set list. f (the_Var (fst p)) = snd p) \<and> 
    (\<forall> \<sigma> \<in> unifiers (set list). \<sigma> = subst_of list \<circ>s \<sigma>) \<and> (subst_domain (subst_of list) \<inter> \<Union> ((vars_term \<circ> snd) ` set eqs) \<subseteq> {}))"])
    case Nil thus ?case by (auto simp: unifiers_def)
  next
    case (Cons p ps)
    show ?case 
    proof
      assume solvepps: "solved (set eqs \<union> set (p # ps))"
      have "set eqs \<union> set ps \<subseteq> set eqs \<union> set (p # ps)" by auto
      from this and solvepps have solveps: "solved (set eqs \<union> set ps)" by (rule solved_mono)
      from Cons[rule_format, OF this] obtain f
        where ind: "subst_of ps = f \<and> (\<forall>p\<in>set ps. f (the_Var (fst p)) = snd p)
        \<and> (\<forall>\<sigma>\<in>unifiers (set ps). \<sigma> = subst_of ps \<circ>s \<sigma>)
        \<and> (subst_domain (subst_of ps) \<inter> \<Union> ((vars_term \<circ> snd) ` set eqs) \<subseteq> {})" by blast
      hence indSub: "subst_of ps = f" and 
        indPs: "\<forall> p \<in> set ps. f (the_Var (fst p)) = snd p" and
        indUn: "\<forall> \<sigma> \<in> unifiers (set ps). \<sigma> = subst_of ps \<circ>s \<sigma>" and
        indDom: "subst_domain (subst_of ps) \<inter> \<Union> ((vars_term \<circ> snd) ` set eqs) \<subseteq> {}"
        by auto
      show "\<exists>f. subst_of (p # ps) = f \<and>
        (\<forall>p\<in>set (p # ps). f (the_Var (fst p)) = snd p) \<and> (\<forall>\<sigma>\<in>unifiers (set (p # ps)). \<sigma> = subst_of (p # ps) \<circ>s \<sigma>) \<and> subst_domain (subst_of (p # ps)) \<inter> \<Union>(vars_term \<circ> snd) ` set eqs \<subseteq> {}"
      proof (cases p)
        case (Pair s t)
        show ?thesis
        proof (cases s)
          case Fun
          with Pair and solvepps have False by (unfold solved_def, auto) thus ?thesis ..
        next
          case (Var x)
          with Pair have p: "p = (Var x, t)" by simp
          let ?g = "(\<lambda>y. if x = y then t else f y) :: 'v \<Rightarrow> ('f,'v) term"
          have "subst_of ((Var x, t) # ps) = ?g"
          proof (simp)
            show "(\<lambda>y. if x = y then t else subst_of ps y) = ?g"
            proof (rule ext)
              fix y
              from indSub show "(if x = y then t else subst_of ps y) = ?g y" by (cases "x = y", auto)
            qed
          qed
          with p have one: "subst_of (p # ps) = ?g" by simp	  
          have two: "\<forall>q\<in>set (p # ps). ?g (the_Var (fst q)) = snd q" 
          proof
            fix q :: "(('f,'v) term \<times> ('f,'v) term)"
            assume qpps: "q \<in> set (p # ps)"
            let ?LL = "(if x = the_Var (fst q) then t else f (the_Var (fst q))) :: ('f,'v) term"
            show "?LL = snd q" 
            proof (cases q)
              case (Pair u v)
              show ?thesis 
              proof (cases "u")
                case Fun with qpps and Pair and solvepps have False by (unfold solved_def, auto) thus ?thesis ..
              next
                case (Var y)
                with Pair have q: "q = (Var y, v)" by simp
                have "(if x = y then t else f y) = v" 
                proof (cases "x = y")
                  case False 
                  with q and p and `q \<in> set (p # ps)` have "q \<in> set ps" by simp
                  with indPs and q have "f y = v" by auto
                  with False show ?thesis by simp
                next
                  case True
                  show ?thesis 
                  proof (cases "t = v") 
                    case True with `x = y` show ?thesis by simp
                  next
                    case False
                    have "p \<in> set (p # ps)" by simp
                    with p and solvepps have "\<exists>z. Var x = Var z \<and> z \<notin> vars_term t \<and> z \<notin> vars_eqs (set eqs \<union> set (p # ps) - {(Var x, t)})" by (unfold solved_def, auto)
                    from this obtain z where "Var x = Var z \<and> z \<notin> vars_term t \<and> z \<notin> vars_eqs (set eqs \<union> set (p # ps) - {(Var x, t)})" ..
                    hence contra: "x \<notin> vars_eqs (set (p # ps) - {(Var x, t)})" by (unfold vars_eqs_def, auto)
                    from False and p and q and `q \<in> set (p # ps)` have qset: "q \<in> set (p # ps) - {(Var x,t)}" by blast
                    from True and q have "x \<in> vars_term(fst q)" by simp 
                    with qset have "x \<in> vars_eqs (set (p # ps) - {(Var x, t)})"
                    by (unfold vars_eqs_def, blast)
                    with contra have False .. thus ?thesis ..
                  qed
                qed		
                with p q show ?thesis by auto
              qed
            qed
          qed
          have three: "\<forall> \<sigma> \<in> unifiers (set (p # ps)). \<sigma> = subst_of (p # ps) \<circ>s \<sigma>" 
          proof
            fix \<sigma> 
            assume "\<sigma> \<in> unifiers (set (p # ps))" 
            with p have "\<sigma> \<in> {\<sigma> . Var x\<cdot>\<sigma> = t\<cdot>\<sigma>} \<inter> unifiers (set ps)" by (simp add: unifiers_insert)
            hence xts: "Var x \<cdot> \<sigma> = t \<cdot> \<sigma>" and su: "\<sigma> \<in> unifiers (set ps)" by auto
            from su and indUn and indSub have ind: "\<sigma> = f \<circ>s \<sigma>" by simp
            show "\<sigma> = subst_of (p # ps) \<circ>s \<sigma>"
            proof -
              from ind have "\<sigma> = f \<circ>s \<sigma>" by simp
              have "?g \<circ>s \<sigma> = \<sigma>"
              proof (simp add: subst_compose_def, rule ext)
                fix y
                show "(if x = y then t else f y) \<cdot> \<sigma> = \<sigma> y"
                  proof (cases "x = y")
                  case False
                  from ind have "f y \<cdot> \<sigma> = \<sigma> y"
                  proof -
                    have "f y \<cdot> \<sigma> = Var y \<cdot> f \<cdot> \<sigma>" by simp
                    also have "\<dots> = Var y \<cdot> f \<circ>s \<sigma>" by simp
                    also have "\<dots> = Var y \<cdot> \<sigma>" using ind by simp
                    also have "\<dots> = \<sigma> y" by simp
                    finally show ?thesis . 
                  qed
                  with False show ?thesis by auto
                next
                  case True
                  with xts show ?thesis by simp 		    
                qed
              qed
              with one show ?thesis by simp
            qed
          qed
          have four: "subst_domain (subst_of (p # ps)) \<inter> \<Union> ((vars_term \<circ> snd) ` set eqs) \<subseteq> {}" 
          proof
            fix y
            assume "y \<in> subst_domain (subst_of (p # ps)) \<inter> \<Union>(vars_term \<circ> snd) ` set eqs"
            hence ypps: "y \<in> subst_domain (subst_of (p # ps))" and yeqs: "y \<in> \<Union> (vars_term \<circ> snd) ` set eqs" by auto
            from ypps and one have "y \<in> subst_domain ?g" by simp
            with p have "y = x \<or> y \<in> subst_domain f"
              by (cases "y = x", auto simp: subst_domain_def)
            with indSub have "y = x \<or> y \<in> subst_domain (subst_of ps)" by simp
            with indDom and yeqs have yx: "y = x" by auto
            from this and yeqs obtain q where qeqs: "q \<in> set eqs \<union> set (p # ps)" and xq: "x \<in> vars_term(snd q)" by force
            from p have  "(Var x, t) \<in> set eqs \<union> set (p # ps)" by simp
            from this and solvepps have "\<exists>z. Var x = Var z \<and> z \<notin> vars_term t \<and> z \<notin> vars_eqs (set eqs \<union> set (p # ps) - {(Var x, t)})" 
              by (unfold solved_def, blast)
            from this have xt: "x \<notin> vars_term t" and xnmem: "x \<notin> vars_eqs (set eqs \<union> set (p # ps) - {(Var x,t)})" by auto
            obtain u v where q: "q = (u,v)" by force
            with xq have xv: "x \<in> vars_term v" by simp
            from q and qeqs have uvpair: "(u,v) \<in> set eqs \<union> set (p # ps)" by simp
            with solvepps have "\<exists>z. u = Var z \<and> z \<notin> vars_term v \<and> z \<notin> vars_eqs (set eqs \<union> set (p # ps) - {(u, v)})" by (unfold solved_def, blast)
            from this obtain z where uz: "u = Var z" and  zv: "z \<notin> vars_term v" by blast
            from zv and q and xq have "x \<noteq> z" by auto
            with uvpair and uz have uvpair2: "(u,v) \<in> set eqs \<union> set (p # ps) - {(Var x, t)}" by blast
            from xv have "x \<in> vars_term(snd (u, v))" by simp
            with uvpair2 have "x \<in> vars_eqs (set eqs \<union> set (p # ps) - {(Var x,t)})" by (unfold vars_eqs_def, blast)
            with xnmem have False ..
            thus "y \<in> {}" ..
          qed	  
          from one two three four show ?thesis by blast
        qed
      qed
    qed
  qed
  with solved obtain f where 
    one: "?S = f" and 
    two: "\<forall>p\<in>set eqs. f (the_Var (fst p)) = snd p" and 
    three: "\<forall>\<sigma>\<in>unifiers (set eqs). \<sigma> = ?S \<circ>s \<sigma>" and 
    four: "subst_domain ?S \<inter>  \<Union> ((vars_term \<circ> snd) ` set eqs) \<subseteq> {}" by force
  show ?thesis
  proof 
    from one and two show "?S \<in> unifiers (set eqs)"
    proof (unfold unifiers_def, auto simp: one)
      fix s t
      assume "(s,t) \<in> set eqs"
      show "s \<cdot> f = t \<cdot> f"
      proof (cases s)
        case Fun with `solved (set eqs)` and `(s,t) \<in> set eqs` 
        have False by (unfold solved_def, auto) 
        thus ?thesis ..
      next
        case (Var x)
        with two and `(s,t) \<in> set eqs` and two have piece1: "s \<cdot> f = t \<cdot> Var" by auto
        from four and `(s,t) \<in> set eqs` have "subst_domain ?S \<inter> vars_term t = {}" by force
        with one have noInter: "vars_term t \<inter> subst_domain f = {}" by auto
        have "finite (vars_term t)" by simp
        from this and noInter have piece2: "f |s (vars_term t) = Var" by (rule subst_restrict_empty_set)
        have piece3: "t \<cdot> f = t \<cdot> (f |s (vars_term t))" by (rule coincidence_lemma)
        from piece1 piece2 piece3  show ?thesis by simp
      qed
    qed
  next    
    from three show "\<forall> \<sigma> \<in> unifiers (set eqs). \<sigma> = subst_of eqs \<circ>s \<sigma>" .    
  qed
qed

lemma solved_single:
  assumes xnt: "x \<notin> vars_term t"
  shows "subst x t \<in> unifiers {(Var x, t)} \<and> (\<forall> \<sigma> \<in> unifiers {(Var x,t)}. \<sigma> = subst x t \<circ>s \<sigma>)"
proof -
  from xnt have solve: "solved (set [(Var x,t)])" (is "solved ?P") by (unfold solved_def vars_eqs_def, auto)
  have eq: "subst x t = subst_of [(Var x,t)]" (is "?XT = ?S") by (unfold subst_def, simp, rule ext, force)
  from solve have "?S \<in> unifiers ?P \<and> (\<forall> \<sigma> \<in> unifiers ?P. \<sigma> = ?S  \<circ>s \<sigma>)"
    by (rule solvedMainProperty)
  with eq show ?thesis by simp
qed
  
lemma subst_sound:
  assumes xnt: "x \<notin> vars_term t"
  shows "unifiers (insert (Var x, t) s) =
    unifiers (insert (Var x,t) ((\<lambda>p. (fst p \<cdot> subst x t, snd p \<cdot> subst x t)) ` s))"
proof -
  let ?xt = "subst x t"
  from xnt have "?xt \<in> unifiers {(Var x, t)} \<and> (\<forall> \<sigma> \<in> unifiers {(Var x,t)}. \<sigma> = ?xt \<circ>s \<sigma>)" by (rule solved_single)
  hence general: "\<forall> \<sigma> \<in> unifiers {(Var x,t)}. \<sigma> = ?xt \<circ>s \<sigma>" ..
  show ?thesis 
  proof (rule set_eqI)
    fix \<theta> 
    have "(\<theta> \<in> unifiers (insert (Var x, t) s)) = (Var x \<cdot> \<theta> = t \<cdot> \<theta> \<and> \<theta> \<in> unifiers s)" by (simp add: unifiers_insert)
    also have "\<dots> = (Var x \<cdot> \<theta> = t \<cdot> \<theta> \<and> ?xt \<circ>s \<theta> \<in> unifiers s)" 
    proof -
      have "Var x \<cdot> \<theta> = t \<cdot> \<theta> \<Longrightarrow> \<theta> \<in> unifiers {(Var x,t)}" by (simp add: unifiers_insert)
      with general have "Var x \<cdot> \<theta> = t \<cdot> \<theta> \<Longrightarrow> \<theta> = ?xt \<circ>s \<theta>" by blast
      thus ?thesis by force
    qed
    also have "\<dots> = (Var x \<cdot> \<theta> = t \<cdot> \<theta> \<and> (\<forall> p \<in> s. fst p \<cdot> ?xt \<circ>s \<theta> = snd p \<cdot> ?xt \<circ>s \<theta>))" by (simp add: unifiers_def)
    also have "\<dots> = (Var x \<cdot> \<theta> = t \<cdot> \<theta> \<and> (\<forall> p \<in> s. fst p \<cdot> ?xt \<cdot> \<theta> = snd p \<cdot> ?xt \<cdot> \<theta>))" by force
    also have "\<dots> = (Var x \<cdot> \<theta> = t \<cdot> \<theta> \<and> (\<forall> q \<in> ((\<lambda> p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) ` s). fst q \<cdot> \<theta> = snd q \<cdot> \<theta>))" by force
    also have "\<dots> = (\<theta> \<in> unifiers (insert (Var x,t) ((\<lambda> p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) ` s)))" by (simp add:  unifiers_def)
    finally show "(\<theta> \<in> unifiers (insert (Var x, t) s)) = (\<theta> \<in> unifiers (insert (Var x, t) ((\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) ` s)))" .
  qed
qed

(* result None shows that there are no unifiers *)
lemma unify_sound1:
  assumes "unify xs ys = None" (is "_ = ?R")
  shows "unifiers (set xs \<union> set ys) = {}" (is ?Q)
proof -
  have "unify xs ys = ?R \<longrightarrow> ?Q"
  proof (induct rule: unify_induct)
    case (Subst x t zs ys)
    let ?xt = "subst x t"
    have "(set (map (\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) (zs @ ys) @ [(Var x, t)]) \<union> set []) = 
      insert (Var x, t) (set (map (\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) (zs @ ys)))" (is "?L = _") by simp
    also have "\<dots> = insert (Var x, t) ((\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) ` (set (zs @ ys)))" (is "_ = ?R") by (auto)
    finally have eq:  "unifiers ?L = unifiers ?R" by simp   
    have "unifiers (set ((Var x, t) # zs) \<union> set ys) = unifiers (insert (Var x, t) (set (zs @ ys)))" by simp
    also have "\<dots> = unifiers (insert (Var x, t) ((\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) ` (set (zs @ ys))))" using `x \<notin> vars_term t` by (rule subst_sound)
    also have "\<dots> = {}" using eq and Subst by simp
    finally show ?case .
  qed auto
  with assms show ?thesis by simp
qed

(* result Some eqs has the same unifiers as input *)
lemma unify_sound2:
  assumes "unify xs ys = Some us" (is "_ = ?R")
  shows "unifiers (set xs \<union> set ys) = unifiers (set us)" (is ?Q)
proof -
  have "unify xs ys = ?R \<longrightarrow> ?Q"
  proof (induct rule: unify_induct)
    case (Subst x t zs ys) 
    let ?xt = "subst x t"
    have "(set (map (\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) (zs @ ys) @ [(Var x, t)]) \<union> set []) = 
      insert (Var x, t) (set (map (\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) (zs @ ys)))" (is "?L = _") by simp
    also have "\<dots> = insert (Var x, t) ((\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) ` (set (zs @ ys)))" (is "_ = ?R") by (auto)
    finally have eq:  "unifiers ?L = unifiers ?R" by simp   
    have "unifiers (set ((Var x, t) # zs) \<union> set ys) = unifiers (insert (Var x, t) (set (zs @ ys)))" by simp
    also have "\<dots> = unifiers (insert (Var x, t) ((\<lambda>p. (fst p \<cdot> ?xt, snd p \<cdot> ?xt)) ` (set (zs @ ys))))" using `x \<notin> vars_term t` by (rule subst_sound)
    also have "\<dots> = unifiers (set us)" using eq and Subst by simp
    finally show ?case .
  qed auto
  with assms show ?thesis by simp
qed

(* solved compution *) 
definition solvedC where
  "solvedC seen todo \<equiv> \<forall>(s,t) \<in> seen. \<exists>x.
    s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> vars_eqs ((seen - {(s,t)}) \<union> todo)"

(* if everything in seen is in solved form, then the result is in solved form *)
lemma unify_sound3:  
  assumes "unify xs ys = Some us" and "solvedC (set ys) (set xs)" 
  shows "solvedC (set us) {}" 
proof -
  have "unify xs ys = Some us \<longrightarrow> (solvedC (set ys) (set xs) \<longrightarrow> solvedC (set us) {})"
  proof (induct rule: unify_induct[where ?xs = "xs" and ?ys = "ys"])
    case (FF f ss g ts zs ys) 
    let ?Y = "set ys"
    let ?F = "{(Fun f ss, Fun g ts)}"
    let ?Z = "set zs"
    let ?ZIP = "set (zip ss ts)"
    have "solvedC ?Y (?F \<union> ?Z) \<Longrightarrow> solvedC ?Y (?ZIP \<union> ?Z)"
    proof (unfold solvedC_def)
      assume ind: "\<forall>(s, t) \<in> ?Y. \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> vars_eqs (?Y - {(s, t)} \<union> (?F \<union> ?Z))"
      show "\<forall>(s, t) \<in> ?Y. \<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> vars_eqs (?Y - {(s, t)} \<union> (?ZIP \<union> ?Z))"
      proof
        fix s t
        assume "(s, t) \<in> ?Y"
        with ind have "\<exists>x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> vars_eqs (?Y - {(s, t)} \<union> (?F \<union> ?Z))" by auto
        from this obtain x where c1: "s = Var x \<and> x \<notin> vars_term t" and xF: "x \<notin> vars_eqs (?Y - {(s, t)} \<union> (?F \<union> ?Z))" by auto
        from xF have easy: "x \<notin> vars_eqs (?Y - {(s,t)}) \<and> x \<notin> vars_eqs ?Z" and f: "x \<notin> vars_term(Fun f ss) \<union> vars_term(Fun g ts)" by auto
        have "x \<notin> vars_eqs ?ZIP"
        proof
          assume "x \<in> vars_eqs ?ZIP"
          hence "x \<in> vars_term(Fun f ss) \<or> x \<in> vars_term(Fun g ts)" by (rule vars_eqs_Fun)
          with f show False by simp
        qed
        with easy have "x \<notin> vars_eqs (?Y - {(s, t)} \<union> (?ZIP \<union> ?Z))" by simp
        with c1 show "\<exists> x. s = Var x \<and> x \<notin> vars_term t \<and> x \<notin> vars_eqs (?Y - {(s, t)} \<union> (?ZIP \<union> ?Z))" by blast
      qed
    qed
    with FF show ?case by auto
  next
    case (Subst x t zs ys)
    note xt = Subst(3)
    let ?xt = "subst x t"
    let ?xts = "\<lambda> s. s \<cdot> ?xt"
    let ?xtp = "\<lambda> p. (?xts (fst p), ?xts (snd p))"
    let ?map = "map ?xtp"
    show ?case
    proof (rule impI, rule Subst(1)[THEN mp])
      assume "solvedC (set ys) (set ((Var x, t) # zs))"
      note solved = this[unfolded solvedC_def, rule_format]
      show "solvedC (set ((Var x, t) # ?map ys))
        (set (?map  zs))" (is "solvedC (set (?xt # ?ys)) (set ?zs)")
        unfolding solvedC_def
      proof
        fix u v
        assume "(u,v) \<in> set (?xt # ?ys)"        
        hence "(u,v) = ?xt \<or> (u,v) \<in> set ?ys" by simp
        thus "\<exists>z. u = Var z \<and>
                z \<notin> vars_term v \<and>
                z \<notin> vars_eqs
                      (set (?xt # ?ys) - {(u, v)} \<union>
                       set ?zs)"
        proof
          assume "(u,v) \<in> set ?ys"
          then obtain u' v' where uv: "(u',v') \<in> set ys" and uu': "u = ?xts u'" and vv': "v = ?xts v'"
            by auto
          from solved[OF uv] obtain z
            where *: "u' = Var z" 
              "z \<notin> vars_term v'" "z \<noteq> x" "z \<notin> vars_term t" "z \<notin> vars_eqs (set ys - {(u', v')})" 
              "z \<notin> vars_eqs (set zs)" by auto
          {
            fix us
            assume "z \<in> vars_eqs (?xtp ` us)"
            then obtain u v where uv: "(u,v) \<in> us \<or> (v,u) \<in> us" "z \<in> vars_term (?xts u)"
              unfolding vars_eqs_def by force
            hence "z \<in> vars_term u" using * unfolding vars_term_subst subst_def 
              by (cases "x \<in> vars_term u", auto)
            with uv have "z \<in> vars_eqs us" unfolding vars_eqs_def by force
          } note unmap = this
          from uu' * have u: "u = Var z" by (simp add: subst_def)
          show "\<exists>z. u = Var z \<and>
                z \<notin> vars_term v \<and>
                z \<notin> vars_eqs
                      (set (?xt # ?ys) - {(u, v)} \<union>
                       set ?zs)"
          proof (intro exI conjI, rule u)
            show "z \<notin> vars_term v" unfolding vv' vars_term_subst using * by (auto simp: subst_def)
          next
            show "z \<notin> vars_eqs (set (?xt # ?map ys) - {(u, v)} \<union> set (?map zs))"
            proof
              assume "z \<in> vars_eqs (set (?xt # ?map ys) - {(u, v)} \<union> set (?map zs))"
              with *(6) unmap[of "set zs"] have "z \<in> vars_eqs (set (?xt # ?map ys) - {(u, v)})" by auto
              with * have "z \<in> vars_eqs (set (?map ys) - {(u, v)})" by (auto simp: vars_eqs_def)
              then obtain p where z: "z \<in> vars_eqs {p}" and p: "p \<in> set (?map ys)" and puv: "p \<noteq> (u,v)" 
                by (auto simp: vars_eqs_def)
              from p obtain q where q: "q \<in> set ys" and p: "p = ?xtp q" by auto
              from z[unfolded p] unmap[of "{q}"] have z: "z \<in> vars_eqs {q}" by auto
              from puv have "q \<noteq> (u',v')" unfolding uu' vv' p by auto
              with q have q: "q \<in> set ys - {(u',v')}" by auto
              with z *(5) show False by (force simp: vars_eqs_def)
            qed
          qed
        next
          assume "(u,v) = ?xt"
          hence id: "u = Var x" "v = t" by auto
          show ?thesis
          proof (rule exI[of _ x], unfold id, intro conjI, simp, rule xt, rule)
            assume "x \<in> vars_eqs (set (?xt # ?ys) - {(Var x, t)} \<union> set ?zs)"
            hence "x \<in> vars_eqs (set (?map (ys @ zs)))" using xt unfolding vars_eqs_def by force
            from vars_eqs_extract[OF this] obtain u where "x \<in> vars_term (?xts u)" by auto
            from this[unfolded vars_term_subst] xt show False by (cases "x \<in> vars_term u", auto simp: subst_def)
          qed
        qed
      qed
    qed
  qed (auto simp: solvedC_def)
  with assms show ?thesis by simp
qed


(* when starting the unification algorithm properly, then the result is in solved form *)	  
lemma unify_sound4: assumes res: "unify eqs [] = Some res" shows "solved (set res)"
proof -
  have "solvedC (set []) (set eqs)" by (unfold solvedC_def, simp)
  with  res have "solvedC (set res) {}" by (rule unify_sound3)
  thus ?thesis by (unfold solved_def solvedC_def, auto)
qed

lemma mgu_complete: "mgu s t = None \<Longrightarrow> unifiers {(s,t)} = {}"
proof -
  assume "mgu s t = None" 
  hence "unify [(s,t)] [] = None" by (cases "unify [(s,t)] []", auto)
  hence "unifiers (set [(s,t)] \<union> set []) = {}" by (rule unify_sound1)
  thus ?thesis by simp
qed

lemma mgu_sound: "mgu s t = Some \<delta> \<Longrightarrow> is_smgu \<delta> {(s,t)}"
proof -
  assume result: "mgu s t = Some \<delta>"
  show ?thesis 
  proof (cases "unify [(s,t)] []")
    case None with result show ?thesis by simp
  next
    case (Some res) 
    hence "unifiers (set [(s,t)] \<union> set []) = unifiers (set res)" by (rule unify_sound2)
    hence unif: "unifiers {(s,t)} = unifiers (set res)" by simp 
    from Some have "solved (set res)" by (rule unify_sound4)
    hence "subst_of res \<in> unifiers (set res) \<and> (\<forall>\<sigma>\<in>unifiers (set res). \<sigma> = subst_of res \<circ>s \<sigma>)" by (rule solvedMainProperty)
    with unif and result and Some have "\<delta> \<in> unifiers {(s,t)} \<and> (\<forall>\<sigma>\<in>unifiers {(s,t)}. \<sigma> = \<delta> \<circ>s \<sigma>)" by force
    thus ?thesis by (unfold is_smgu_def, force)
  qed
qed 

lemma the_mgu[simp]:
  fixes \<sigma> :: "('f, 'v) subst"
  assumes unif: "s \<cdot> \<sigma> = t \<cdot> \<sigma>"
  shows "s \<cdot> the_mgu s t = t \<cdot> the_mgu s t \<and> \<sigma> = the_mgu s t \<circ>s \<sigma>" 
proof -
  from unif have ne: "\<sigma> \<in> unifiers {(s,t)}" by (unfold unifiers_def, force)
  show ?thesis
  proof (cases "mgu s t")
    assume "mgu s t = None"
    hence "unifiers {(s,t)} = {}" by (rule mgu_complete)
    with ne have False by simp thus ?thesis ..
  next
    fix \<delta> 
    assume res: "mgu s t = Some \<delta>"
    hence "is_smgu \<delta> {(s,t)}" by (rule mgu_sound)
    with res have "is_smgu (the_mgu s t) {(s,t)}" by (unfold the_mgu_def, simp)
    with ne show ?thesis by (unfold is_smgu_def unifiers_def, auto)
  qed
qed

hide_const(open) solved

section {* Matching *}

definition matcher :: "('f, 'w) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v, 'w) gsubst set" where
  "matcher t l \<equiv> {\<sigma>. t = l \<cdot> \<sigma>}"

definition matcherL :: "(('f, 'w) term \<times> ('f, 'v) term) set \<Rightarrow> ('f, 'v,'w) gsubst set" where
  "matcherL ps \<equiv> {\<sigma>. (\<forall>p \<in> ps. fst p = snd p \<cdot> \<sigma>)}"

abbreviation matcherLi :: "('v \<times> ('f, 'w) term) set \<Rightarrow> ('f, 'v,'w) gsubst set" where
  "matcherLi ps \<equiv> matcherL ((\<lambda>p. (snd p, Var (fst p))) ` ps)"


(* a sound and complete matching algorithm *)

function
  match_term_list :: "(('f, 'w) term \<times> ('f, 'v) term) list \<Rightarrow> ('v \<times> ('f, 'w) term) list option"
where
  "match_term_list [] = Some []"
| "match_term_list ((Var _, Fun _ _) # ps) = None"
| "match_term_list ((Fun f ts,Fun g ls) # ps) = do {
    guard (f = g);
    eqs \<leftarrow> zip_option ts ls;
    match_term_list (eqs @ ps)
  }"
| "match_term_list ((t, Var x) # ps) = do {
    eqs \<leftarrow> match_term_list ps;
    Some ((x, t) # eqs)
  }"
  by pat_completeness auto

termination 
proof -
  let ?M = "measures [\<lambda>l. list_size (size \<circ> snd) l, length]"
  show ?thesis 
  proof
    fix f::'f and ts :: "('f, 'w) term list" and g and ls :: "('f, 'v) term list" and ps res
    assume "f = g" and dec: "zip_option ts ls = Some res"
    from dec have len: "length ts = length ls" by auto
    from dec have res: "res = zip ts ls" by simp
    from len and res have main: "list_size (size \<circ> snd) (res @ ps) < Suc (Suc (list_size size ls + list_size (size \<circ> snd) ps))"
      by (auto simp: list_size_map[symmetric] simp del: list_size_map)
    show "(res @ ps, (Fun f ts, Fun g ls) # ps) \<in> ?M"
      by (rule measures_less, auto simp: main simp del: list_size_append)
  qed auto
qed

lemma matcherL_Un[simp]: "matcherL (ps \<union> qs) = matcherL ps \<inter> matcherL qs"
  by (unfold matcherL_def, blast)

lemma matcherL_insert: "matcherL (insert p ps) = {\<sigma>. fst p = snd p \<cdot> \<sigma>} \<inter> matcherL ps"
  by (unfold matcherL_def, blast)

lemma matcherL_zip: assumes "length ts = length ls" shows 
  "{\<sigma>. \<forall>i<length ts. ts ! i = ls ! i \<cdot> \<sigma>} = matcherL (set (zip ts ls))" 
using assms
proof (induct ls arbitrary:ts) 
  case Nil thus ?case by (cases ts, auto simp:matcherL_def)
next
  case (Cons l ls) note oCons=this
  thus ?case 
  proof (cases ts)
    case Nil with Cons show ?thesis by simp
  next
    case (Cons t tts)
    from Cons oCons have "length tts = length ls" by simp
    with Cons oCons have Prop: "{\<sigma>. \<forall>i<length tts. tts ! i = ls ! i \<cdot> \<sigma>} = matcherL (set (zip tts ls))" by simp
    have "{\<sigma>. \<forall>i<length (t # tts). (t # tts) ! i = (l # ls) ! i \<cdot> \<sigma>} = { \<sigma>. t = l \<cdot> \<sigma>} \<inter> {\<sigma>. \<forall>i<length tts. tts ! i = ls ! i \<cdot> \<sigma>}" (is "?l = ?r")
    proof (intro set_eqI iffI)
      fix s
      assume "s \<in> ?l"
      thus "s  \<in> ?r" by force
    next
      fix s 
      assume Prop2: "s \<in> ?r"
      hence one: "t = l \<cdot> s" by simp
      from Prop2 have two: "\<forall>i< length tts. tts ! i = ls ! i \<cdot> s" by simp
      have "\<forall> i < length (t # tts). (t # tts) ! i = (l # ls) ! i \<cdot> s"
      proof (intro allI impI)
        fix i
        assume "i < length (t # tts)"
        with one two show "(t # tts) ! i = (l # ls) ! i \<cdot> s"
          by (cases i, auto)
      qed
      thus "s \<in> ?l" by simp
    qed
    also have "\<dots> = { \<sigma>. t = l \<cdot> \<sigma>} \<inter> matcherL (set (zip tts ls))" using Prop by simp
    finally show ?thesis using Cons by (simp add: matcherL_insert)
  qed
qed

lemma matcherL_Fun[simp]: "matcherL (insert (Fun f ts, Fun g ls) ps) = (if f = g \<and> length ts = length ls then matcherL (set (zip ts ls)) \<inter> matcherL ps else
{})" 
proof (cases "f = g \<and> length ts = length ls")
  assume "\<not> (f = g \<and> length ts = length ls)"
  hence "f \<noteq> g \<or> length ts \<noteq> length ls" by blast
  thus ?thesis
  proof
    assume "f \<noteq> g" thus ?thesis by (unfold matcherL_def, simp)
  next
    assume "length ts \<noteq> length ls" thus ?thesis by (unfold matcherL_def, auto)
  qed
next
  assume P: "f = g \<and> length ts = length ls"
  hence "{\<sigma>. ts = map (\<lambda> l. l \<cdot> \<sigma>) ls} = {\<sigma>. \<forall>i<length ts. ts ! i = (map (\<lambda> l. l \<cdot> \<sigma>) ls) ! i}"  using list_eq_iff_nth_eq[where xs = ts] by force
  also have "\<dots> = {\<sigma>. \<forall>i<length ts. ts ! i = ls ! i \<cdot> \<sigma>}" using P by auto 
  also have "\<dots> = matcherL (set (zip ts ls))"
    using P matcherL_zip by blast
  finally show ?thesis using P by (simp add: matcherL_insert)
qed

lemma match_term_list_complete:
  assumes "match_term_list ps = None" shows "matcherL (set ps) = {}"
using assms 
proof (induct rule:match_term_list.induct)
  case 1 thus ?case by simp
next
  case (2 x f ls ps) thus ?case by (unfold matcherL_def, auto)
next
  case (4 t x ps) thus ?case by (cases "match_term_list ps", unfold matcherL_def, auto)
next
  case (3 f ts g ls ps) thus ?case 
  proof (cases "f = g")
    case False thus ?thesis by (unfold matcherL_def, auto)
  next
    case True thus ?thesis
    proof (cases "zip_option ts ls")
      case (Some eqs)
      hence len: "length ts = length ls" by auto
      from Some have zip: "eqs = zip ts ls" by auto
      with Some `f = g` 3(1)[of eqs] 3(2)
        have "matcherL (set (zip ts ls @ ps)) = {}" by simp
      with len show ?thesis by simp
    next
      case None
      hence difflen: "length ts \<noteq> length ls" by simp
      thus ?thesis by simp
    qed
  qed
qed

lemma match_term_list_sound:
  assumes "match_term_list ps = Some qs" shows "matcherL (set ps) = matcherLi (set qs)" 
using assms 
proof (induct arbitrary:qs rule:match_term_list.induct)
  case 1 thus ?case by (unfold matcherL_def, simp)
next
  case (2 x f ls ps) thus ?case by auto
next
  case (3 f ts g ls ps) thus ?case 
  proof (cases "f = g")
    case False with 3 show ?thesis by auto
  next
    case True thus ?thesis
    proof (cases "zip_option ts ls")
      case None with 3 and `f=g` show ?thesis by simp
    next
      case (Some eqs)
      hence len: "length ts = length ls" by auto
      from Some have zip: "eqs = zip ts ls" by auto
      with 3 Some True zip have "matcherL (set (zip ts ls @ ps)) = matcherLi (set qs)" by simp
      with len True show ?thesis by simp
    qed
  qed
next
  case (4 t x ps) thus ?case 
  proof (cases "match_term_list ps")
    case None with 4 show ?thesis by simp
  next
    case (Some eqs) with 4 show ?thesis by (auto simp: matcherL_insert)
  qed
qed

abbreviation rvars :: "(('f, 'w)term \<times> ('f,'v)term) list \<Rightarrow> 'v set" where
  "rvars ps \<equiv> \<Union> ((vars_term \<circ> snd) ` (set ps))"

abbreviation lvars :: "('v \<times> ('f, 'w) term) list \<Rightarrow> 'v set" where
  "lvars ps \<equiv> { x | t x. (x,t) \<in> set ps}"

lemma match_term_list_vars: assumes "match_term_list ps = Some qs" shows "rvars ps = lvars qs"
using assms
proof (induct arbitrary:qs rule:match_term_list.induct)
  case 1 thus ?case by auto
next
  case (2 x f ls ps) thus ?case by auto
next
  case (4 t x ps) thus ?case by (cases "match_term_list ps", auto)
next
  case (3 f ts g ls ps) thus ?case 
  proof (cases "f = g")
    case False with 3 show ?thesis by auto
  next
    case True thus ?thesis
    proof (cases "zip_option ts ls")
      case None with 3 and `f=g` show ?thesis by simp
    next
      case (Some eqs)
      hence len: "length ts = length ls" by auto
      from Some have zip: "eqs = zip ts ls" by auto
      with 3 Some True zip have ind: "rvars (eqs @ ps) = lvars qs" by auto
      from len have "rvars (zip ts ls) = rvars [(Fun f ts, Fun g ls)]" 
      proof (induct ls arbitrary: ts, simp)
        case (Cons l ls)
        thus ?case by (cases ts, auto)
      qed
      with ind zip show "rvars ((Fun f ts, Fun g ls) # ps) = lvars qs" by auto
    qed
  qed
qed

fun
  match_vars :: "('v \<times> ('f, 'w) term) list \<Rightarrow> ('f, 'v, 'w) gsubstL option"
where
  "match_vars [] = Some []"
| "match_vars ((x, t) # ps) =
    (case map_of ps x of 
      None \<Rightarrow>
      (case match_vars ps of
        Some \<sigma> \<Rightarrow> Some ((x, t) # \<sigma>)
      | None \<Rightarrow> None)
    | Some s \<Rightarrow> if t = s then match_vars ps else None)"

lemma match_vars_sound:
  assumes "match_vars ps = Some \<sigma>"
  shows "mk_subst d \<sigma> \<in> matcherLi (set ps)"
proof -
  {
    assume "match_vars ps = Some \<sigma>"
    hence "mk_subst d \<sigma> \<in> matcherLi (set ps)" 
    proof (induct ps arbitrary: \<sigma>, simp add: matcherL_def)
      case (Cons xt qs)
      show ?case 
      proof (cases xt)
        case (Pair x t)
        show ?thesis 
        proof (cases "map_of qs x")
          case (Some s)
          show ?thesis 
          proof (cases "s = t")
            case False
            with Some Pair Cons show ?thesis by simp
          next
            case True
            from Some have "(x,s) \<in> set qs" by (rule map_of_SomeD)
            with True Some Pair have "set (Cons xt qs) = set qs" by auto
            with Pair Cons True Some show ?thesis by auto
          qed
        next
          case None
          with Cons Pair show ?thesis 	    
          proof (cases "match_vars qs", simp)
            case (Some \<delta>)
            show ?thesis
            proof -
            from Cons Pair None Some
              have id: "mk_subst d \<sigma> = mk_subst d ((x, t)#\<delta>)" by auto
            with Cons Some have ind: "mk_subst d \<delta> \<in> matcherLi (set qs)" by auto
	          from None have "\<not> (\<exists> t. (x,t) \<in> set qs)" by (force simp: map_of_eq_None_iff)
	          with ind Pair show ?thesis
            unfolding matcherL_def id by (auto simp add: mk_subst_def)
	        qed
	      qed
	    qed
    qed
  qed
  }
  with assms show ?thesis by auto
qed

lemma match_vars_complete:
  assumes "\<delta> \<in> matcherLi (set ps)" 
  shows "\<exists>\<sigma>. match_vars ps = Some \<sigma> \<and> (\<forall>x \<in> lvars ps. \<delta> x = mk_subst d \<sigma> x)" 
proof -
  {
  let ?prop = "\<lambda>\<sigma> ps. match_vars ps = Some \<sigma> \<and> (\<forall> x \<in> lvars ps. \<delta> x = mk_subst d \<sigma> x)"
  assume "\<delta> \<in> matcherLi (set ps)"
  hence "\<exists>\<sigma>. ?prop \<sigma> ps"
  proof (induct ps)
    case Nil thus ?case by auto
  next
    case (Cons xt qs)
    show ?case 
    proof (cases xt)
      case (Pair x t)
      with Cons have delt: "\<delta> x = t" unfolding matcherL_def by auto
      from Cons have "\<delta> \<in> matcherLi (set qs)" unfolding matcherL_def by auto 
      with Cons obtain \<sigma> where ind: "?prop \<sigma> qs" by force
      show ?thesis 
      proof (cases "map_of qs x")
        case (Some s)
        hence  "(x,s) \<in> set qs" by (rule map_of_SomeD)
        with Cons have dels: "\<delta> x = s" unfolding matcherL_def by auto
        with ind `(x,s) \<in> set qs` have "\<delta> x = mk_subst d \<sigma> x" by auto
        with dels delt ind Pair Some have "?prop \<sigma> (xt # qs)"  by auto
        thus ?thesis by blast
      next
        case None
        hence "\<not> (\<exists> s. (x,s) \<in> set qs)" by (force simp: map_of_eq_None_iff)
        have "?prop ((x, t)#\<sigma>) (xt # qs)" using ind Pair None delt by (auto simp: mk_subst_def)
        thus ?thesis by blast
      qed
    qed
  qed
  }
  with assms show ?thesis by auto
qed

fun
  match_list :: "(('f, 'w) term \<times> ('f, 'v) term) list \<Rightarrow> ('v \<times> ('f, 'w) term) list option"
where
  "match_list tls =
    (case match_term_list tls of 
      None \<Rightarrow> None 
    | Some ps \<Rightarrow> match_vars ps)"

fun match :: "('f, 'w) term \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v, 'w) gsubst option" where
  "match t l = (case match_list [(t, l)] of
    None \<Rightarrow> None
  | Some bs \<Rightarrow> Some (mk_subst (\<lambda>x. t) bs))"

lemma match_list_sound:
  assumes match: "match_list tls = Some \<sigma>"
  shows "\<And>t l. (t, l) \<in> set tls \<Longrightarrow> l \<cdot> mk_subst d \<sigma> = t"
proof -
  fix t l
  assume tl: "(t,l) \<in> set tls"
  show "l \<cdot> mk_subst d \<sigma> = t"
    using match
  proof (cases "match_term_list tls", simp)
    case (Some ps) note oSome = this
    thus ?thesis using match 
    proof (cases "match_vars ps", simp)
      case (Some \<delta>)
      with match_vars_sound have "mk_subst d \<delta> \<in> matcherLi (set ps)" by auto
      with match_term_list_sound oSome have "mk_subst d \<delta> \<in> matcherL (set tls)" by force
      with tl oSome Some match show ?thesis unfolding matcherL_def by auto
    qed
  qed
qed

lemma match_sound:
  assumes "match t l = Some \<sigma>" shows "l \<cdot> \<sigma> = t"
proof -
  from assms obtain bs where "match_list [(t, l)] = Some bs"
    and "mk_subst (\<lambda>x. t) bs = \<sigma>" by (auto split: option.splits)
  from match_list_sound [OF this(1), of t l] and this(2)
    show ?thesis by force
qed

lemma match_list_complete:
  assumes "\<And> t l. (t, l) \<in> set tls \<Longrightarrow> l \<cdot> \<delta> = t" 
  shows "\<exists> \<sigma> . match_list tls = Some \<sigma> \<and> (\<forall> x \<in> rvars tls. \<delta> x = mk_subst d \<sigma> x)"
proof -
  from assms have match: "\<delta> \<in> matcherL (set tls)" unfolding matcherL_def by auto
  show ?thesis
  proof (cases "match_term_list tls")
    case None
    hence "matcherL (set tls) = {}" by (rule match_term_list_complete)
    with match show ?thesis by simp
  next
    case (Some ps)
    hence "matcherL (set tls) = matcherLi (set ps)" by (rule match_term_list_sound)
    with match have "\<delta> \<in> matcherLi (set ps)" by auto
    hence "\<exists> \<sigma>. match_vars ps = Some \<sigma> \<and>  (\<forall>x\<in>lvars ps. \<delta> x = mk_subst d \<sigma> x)"
      by (rule match_vars_complete)
    from this obtain \<sigma> where one: "match_vars ps = Some \<sigma> \<and>  (\<forall>x\<in>lvars ps. \<delta> x = mk_subst d \<sigma> x)" ..
    from Some have "rvars tls = lvars ps" by (rule match_term_list_vars)
    with Some one have "match_list tls = Some \<sigma> \<and> (\<forall>x \<in> rvars tls. \<delta> x = mk_subst d \<sigma> x)" by auto
    thus ?thesis by force
  qed
qed

lemma match_complete:
  assumes "l \<cdot> \<delta> = t" 
  shows "\<exists>\<sigma>. match t l = Some \<sigma> \<and> (\<forall> x \<in> vars_term l. \<delta> x = \<sigma> x)"
proof -
  from assms have precond: "\<And> t' l'. (t',l') \<in> set [(t, l)] \<Longrightarrow> l' \<cdot> \<delta> = t'" by simp
  from match_list_complete[OF precond]
  obtain \<sigma> where "match_list [(t, l)] = Some \<sigma>"
    and "\<forall> x \<in> rvars [(t, l)]. \<delta> x = mk_subst (\<lambda>x. t) \<sigma> x" by blast
  thus ?thesis by auto
qed
      
declare match.simps[simp del]

definition matches :: "('f, 'w) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" where
  "matches t l \<equiv> case match t l of None \<Rightarrow> False | Some _ \<Rightarrow> True"

lemma matches [simp]: "matches t l = (\<exists>\<sigma> . t = l \<cdot> \<sigma>)"
proof (cases "match t l")
  case (Some \<sigma>)
  with match_sound[OF Some] show ?thesis unfolding matches_def by auto
next
  case None
  have "\<not> (\<exists> \<sigma>. t = l \<cdot> \<sigma>)"
  proof (rule notI)
    assume "\<exists> \<sigma> . t = l \<cdot> \<sigma>"
    then obtain \<sigma> where t: "l \<cdot> \<sigma> = t" by auto
    from match_complete[OF t] None show False by auto
  qed
  with None show ?thesis unfolding matches_def by auto
qed

(* checking \<exists> \<sigma> \<delta>. s \<sigma> = t \<delta> via mgu's and variable renamings *)
(* important note: given s and t :: ('f,'v)term, 
   of course, it would be convinient if one can just
   internally switch to pairs and then afterwards present mgu's \<mu>1
   and \<mu>2 for lhs and rhs of type ('f,'v)subst with result
   \<exists>\<tau>1. \<sigma> = \<mu>1 \<circ>s \<tau>1 and \<exists>\<tau>2. \<delta> = \<mu>2 \<circ>s \<tau>2.
   However, this is not possible for arbitrary 'v: consider the
   example where UNIV :: 'v set = {x,y,z}, s = f(x,g(y,z)), t = f(g(y,z),x)
   a mgu \<mu>1 for the lhs s obviously must replace x/g(_,_) y/_, z/_ where each _ 
   is a variable (because a solution is \<sigma> = \<tau> = {x/g(y,z)}).
   However, there is also a solution \<sigma> = {x/g(a,b), y/c, z/d} and \<tau> = {y/a,z/b,x/g(a,b)}
   hence, all four variables _ in the rhss of \<mu>1 must be different.
   But since the universe consists only of three variables, this is not possible!

   Conclusion: to unify var disjoint one definitely requires a way to generate new variables,
      since the variables of the terms do not suffice to express the unifier of even one side.
   This can be done by enlarging the type (as done in the following) or by some fresh name generator.
*)
definition
  mgu_var_disjoint_generic :: "('v \<Rightarrow> 'u) \<Rightarrow> ('w \<Rightarrow> 'u) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'w) term \<Rightarrow> (('f,'v,'u)gsubst \<times> ('f, 'w,'u)gsubst) option"
where
  "mgu_var_disjoint_generic vu wu s t \<equiv> case mgu (map_vars vu s) (map_vars wu t) of
     None \<Rightarrow> None 
   | Some \<gamma> \<Rightarrow> Some (map_subst_domain vu \<gamma>, map_subst_domain wu \<gamma>)"

lemma mgu_var_disjoint_generic_sound: 
  assumes unif: "mgu_var_disjoint_generic vu wu s t = Some (\<gamma>1,\<gamma>2)"
  shows "s \<cdot> \<gamma>1 = t \<cdot> \<gamma>2"
proof -
  from unif[unfolded mgu_var_disjoint_generic_def] obtain \<gamma> where
    unif2: "mgu (map_vars vu s) (map_vars wu t) = Some \<gamma>"
    by (cases "mgu (map_vars vu s) (map_vars wu t)", auto)
  from mgu_sound[OF unif2[unfolded mgu_var_disjoint_generic_def], unfolded is_smgu_def unifiers_def]
  have "map_vars vu s \<cdot> \<gamma> = map_vars wu t \<cdot> \<gamma>" by auto
  from this[unfolded map_subst_domain] unif[unfolded mgu_var_disjoint_generic_def unif2] 
  show ?thesis by simp
qed

(* if terms s and t can become identical via two substitutions \<sigma> and \<delta> 
   then mgu_var_disjoint yields two more general substitutions \<mu>1 \<mu>2 *)
lemma mgu_var_disjoint_generic_complete:
  fixes \<sigma> :: "('f, 'v, 'u) gsubst" and \<tau> :: "('f,'w,'u)gsubst" 
  and vu :: "'v \<Rightarrow> 'u" and wu:: "'w \<Rightarrow> 'u"
  assumes inj: "inj vu" "inj wu"
  and vwu: "range vu \<inter> range wu = {}"
  and unif_disj: "s \<cdot> \<sigma> = t \<cdot> \<tau>"
  shows "\<exists>\<mu>1 \<mu>2 \<delta>. mgu_var_disjoint_generic vu wu s t = Some (\<mu>1,\<mu>2) \<and> 
    \<sigma> = \<mu>1 \<circ>s \<delta> \<and>
    \<tau> = \<mu>2 \<circ>s \<delta> \<and>
    s \<cdot> \<mu>1 = t \<cdot> \<mu>2"
proof -
  note inv1[simp] = the_inv_f_f[OF inj(1)]
  note inv2[simp] = the_inv_f_f[OF inj(2)]
  obtain \<gamma> :: "('f,'u)subst" where gamma: "\<gamma> = (\<lambda> x. if x \<in> range vu then \<sigma> (the_inv vu x) else \<tau> (the_inv wu x))" by auto 
  have ids: "s \<cdot> \<sigma> = map_vars vu s \<cdot> \<gamma>" unfolding gamma
    by (induct s, auto)
  have idt: "t \<cdot> \<tau> = map_vars wu t \<cdot> \<gamma>" unfolding gamma
    by (induct t, insert vwu, auto)
  from unif_disj ids idt
  have unif: "map_vars vu s \<cdot> \<gamma> = map_vars wu t \<cdot> \<gamma>" (is "?s \<cdot> \<gamma> = ?t \<cdot> \<gamma>") by auto
  from the_mgu[OF unif] have unif2: "?s \<cdot> the_mgu ?s ?t = ?t \<cdot> the_mgu ?s ?t" and inst: "\<gamma> = the_mgu ?s ?t \<circ>s \<gamma>" by auto
  have "mgu ?s ?t = Some (the_mgu ?s ?t)" unfolding the_mgu_def
    using mgu_complete[unfolded unifiers_def] unif
    by (cases "mgu ?s ?t", auto)
  with inst obtain \<mu> where mu: "mgu ?s ?t = Some \<mu>" and gamma_mu: "\<gamma> = \<mu> \<circ>s \<gamma>" by auto
  let ?tau1 = "map_subst_domain vu \<mu>" 
  let ?tau2 = "map_subst_domain wu \<mu>" 
  show ?thesis unfolding mgu_var_disjoint_generic_def mu option.simps
  proof (intro exI conjI, rule refl)
    show "\<sigma> = ?tau1 \<circ>s \<gamma>"
    proof (rule ext)
      fix x
      have "(?tau1 \<circ>s \<gamma>) x = \<gamma> (vu x)" using fun_cong[OF gamma_mu, of "vu x"] by (simp add: subst_compose_def)
      also have "... = \<sigma> x" unfolding gamma by simp
      finally show "\<sigma> x = (?tau1 \<circ>s \<gamma>) x" by simp
    qed
  next
    show "\<tau> = ?tau2 \<circ>s \<gamma>"
    proof (rule ext)
      fix x
      have "(?tau2 \<circ>s \<gamma>) x = \<gamma> (wu x)" using fun_cong[OF gamma_mu, of "wu x"] by (simp add: subst_compose_def)
      also have "... = \<tau> x" unfolding gamma using vwu by auto
      finally show "\<tau> x = (?tau2 \<circ>s \<gamma>) x" by simp
    qed
  next
    have "s \<cdot> ?tau1 = map_vars vu s \<cdot> \<mu>" unfolding map_subst_domain ..
    also have "... = map_vars wu t \<cdot> \<mu>"
      unfolding unif2[unfolded the_mgu_def mu option.simps] ..
    also have "... = t \<cdot> ?tau2" unfolding map_subst_domain ..
    finally show "s \<cdot> ?tau1 = t \<cdot> ?tau2" .
  qed
qed

abbreviation mgu_var_disjoint_pair where "mgu_var_disjoint_pair \<equiv> mgu_var_disjoint_generic Inl Inr"


lemma mgu_var_disjoint_pair_sound: 
  "mgu_var_disjoint_pair s t = Some (\<gamma>1,\<gamma>2) \<Longrightarrow> s \<cdot> \<gamma>1 = t \<cdot> \<gamma>2"
  by (rule mgu_var_disjoint_generic_sound)

lemma mgu_var_disjoint_pair_complete:
  fixes \<sigma> :: "('f, 'v, 'v + 'w) gsubst" and \<tau> :: "('f,'w,'v + 'w)gsubst" 
  assumes unif_disj: "s \<cdot> \<sigma> = t \<cdot> \<tau>"
  shows "\<exists>\<mu>1 \<mu>2 \<delta>. mgu_var_disjoint_pair s t = Some (\<mu>1,\<mu>2) \<and> 
    \<sigma> = \<mu>1 \<circ>s \<delta> \<and>
    \<tau> = \<mu>2 \<circ>s \<delta> \<and>
    s \<cdot> \<mu>1 = t \<cdot> \<mu>2"
  by (rule mgu_var_disjoint_generic_complete[OF _ _ _ unif_disj], auto simp: inj_on_def)

lemma mgu_var_disjoint_pair_instance:
  fixes \<sigma> :: "('f, 'v)subst" and \<delta> :: "('f,'v)subst"
  assumes unif_disj: "s \<cdot> \<sigma> = t \<cdot> \<delta>"
  shows "\<exists>\<mu>1 \<mu>2 \<tau>. mgu_var_disjoint_pair s t = Some (\<mu>1,\<mu>2) \<and> 
    \<sigma> = \<mu>1 \<circ>s \<tau> \<and>
    \<delta> = \<mu>2 \<circ>s \<tau> \<and> 
    s \<cdot> \<mu>1 = t \<cdot> \<mu>2"
proof -
  let ?m = "map_vars_subst_ran (Inl :: ('v \<Rightarrow> 'v + 'v))"
  let ?m' = "map_vars_subst_ran (sum_case (\<lambda> x. x) (\<lambda> x. x))"
  from unif_disj have id: "map_vars Inl (s \<cdot> \<sigma>) = map_vars Inl (t \<cdot> \<delta>)" by simp
  from mgu_var_disjoint_pair_complete[OF id[unfolded map_vars_subst_ran]]
  obtain \<mu>1 \<mu>2 \<tau> where mgu: "mgu_var_disjoint_pair s t = Some (\<mu>1,\<mu>2)"
    and \<sigma>: "?m \<sigma> = \<mu>1 \<circ>s \<tau>" 
    and \<delta>: "?m \<delta> = \<mu>2 \<circ>s \<tau>"
    and unif: "s \<cdot> \<mu>1 = t \<cdot> \<mu>2" by blast
  {
    fix \<sigma> :: "('f,'v)subst"
    have "?m' (?m \<sigma>) = \<sigma>" by (simp add: map_vars_compose o_def)
  } note id = this
  {
    fix \<mu> :: "('f,'v,'v+'v)gsubst" and \<tau> :: "('f,'v + 'v)subst"
    have "?m' (\<mu> \<circ>s \<tau>) = \<mu> \<circ>s ?m' \<tau>"
      by (rule ext, unfold subst_compose_def, simp add: map_vars_subst_ran)
  } note id' = this
  from arg_cong[OF \<sigma>, of ?m', unfolded id id'] have \<sigma>: "\<sigma> = \<mu>1 \<circ>s ?m' \<tau>" .
  from arg_cong[OF \<delta>, of ?m', unfolded id id'] have \<delta>: "\<delta> = \<mu>2 \<circ>s ?m' \<tau>" .
  show ?thesis
    by (intro exI conjI, rule mgu, rule \<sigma>, rule \<delta>, rule unif)
qed


abbreviation (input) x_var :: "string \<Rightarrow> string" where "x_var \<equiv> Cons (CHR ''x'')"
abbreviation (input) y_var :: "string \<Rightarrow> string" where "y_var \<equiv> Cons (CHR ''y'')"
abbreviation (input) z_var :: "string \<Rightarrow> string" where "z_var \<equiv> Cons (CHR ''z'')"

abbreviation mgu_var_disjoint_string where "mgu_var_disjoint_string \<equiv> mgu_var_disjoint_generic x_var y_var"

lemma mgu_var_disjoint_string_sound: 
  "mgu_var_disjoint_string s t = Some (\<gamma>1,\<gamma>2) \<Longrightarrow> s \<cdot> \<gamma>1 = t \<cdot> \<gamma>2"
  by (rule mgu_var_disjoint_generic_sound)

lemma mgu_var_disjoint_string_complete:
  fixes \<sigma> :: "('f, string)subst" and \<tau> :: "('f,string)subst" 
  assumes unif_disj: "s \<cdot> \<sigma> = t \<cdot> \<tau>"
  shows "\<exists>\<mu>1 \<mu>2 \<delta>. mgu_var_disjoint_string s t = Some (\<mu>1,\<mu>2) \<and> 
    \<sigma> = \<mu>1 \<circ>s \<delta> \<and>
    \<tau> = \<mu>2 \<circ>s \<delta> \<and>
    s \<cdot> \<mu>1 = t \<cdot> \<mu>2"
  by (rule mgu_var_disjoint_generic_complete[OF _ _ _ unif_disj], auto simp: inj_on_def)


lemma mgu_var_disjoint_right: fixes s t :: "('f,'v)term" and \<sigma> \<tau> :: "('f,'v)subst" and T 
  assumes s: "vars_term s \<subseteq> S"
  and inj: "inj T"
  and ST: "S \<inter> range T = {}"
  and id: "s \<cdot> \<sigma> = t \<cdot> \<tau>"
  shows "\<exists> \<mu> \<delta>. mgu s (map_vars T t) = Some \<mu> \<and> s \<cdot> \<sigma> = s \<cdot> \<mu> \<cdot> \<delta> \<and> (\<forall> t :: ('f,'v)term. t \<cdot> \<tau> = map_vars T t \<cdot> \<mu> \<cdot> \<delta>) \<and> (\<forall>x \<in> S. \<sigma> x = \<mu> x \<cdot> \<delta>)"
proof -
  let ?\<sigma> = "\<lambda> x. if x \<in> S then \<sigma> x else \<tau> ((the_inv T) x)"
  let ?t = "map_vars T t"
  have ids: "s \<cdot> \<sigma> = s \<cdot> ?\<sigma>"
    by (rule term_subst_eq, insert s, auto)
  have "t \<cdot> \<tau> = map_vars (the_inv T) ?t \<cdot> \<tau>"
    unfolding map_vars_compose o_def using the_inv_f_f[OF inj] by auto
  also have "... = ?t \<cdot> (map_subst_domain (the_inv T) \<tau>)" unfolding map_subst_domain ..
  also have "... = ?t \<cdot> ?\<sigma>"
  proof (rule term_subst_eq)
    fix x
    assume "x \<in> vars_term ?t"
    hence "x \<in> T ` UNIV" unfolding vars_term_map_vars by auto
    hence "x \<notin> S" using ST by auto
    thus "map_subst_domain (the_inv T) \<tau> x = ?\<sigma> x" by simp
  qed
  finally have idt: "t \<cdot> \<tau> = ?t \<cdot> ?\<sigma>" by simp
  from id[unfolded ids idt] have id: "s \<cdot> ?\<sigma> = ?t \<cdot> ?\<sigma>" .
  with mgu_complete[of s ?t] id obtain \<mu> where \<mu>: "mgu s ?t = Some \<mu>"
    unfolding unifiers_def  by (cases "mgu s ?t", auto)
  from the_mgu[OF id] have id: "s \<cdot> \<mu> = map_vars T t \<cdot> \<mu>" and \<sigma>: "?\<sigma> = \<mu> \<circ>s ?\<sigma>"
    unfolding the_mgu_def \<mu> by auto
  have "s \<cdot> \<sigma> = s \<cdot> (\<mu> \<circ>s ?\<sigma>)" unfolding ids using \<sigma> by simp
  also have "... = s \<cdot> \<mu> \<cdot> ?\<sigma>" by simp
  finally have ids: "s \<cdot> \<sigma> = s \<cdot> \<mu> \<cdot> ?\<sigma>" .
  {
    fix x
    have "\<tau> x = ?\<sigma> (T x)" using ST unfolding the_inv_f_f[OF inj] by auto
    also have "... = (\<mu> \<circ>s ?\<sigma>) (T x)" using \<sigma> by simp
    also have "... = \<mu> (T x) \<cdot> ?\<sigma>" unfolding subst_compose_def by simp
    finally have "\<tau> x = \<mu> (T x) \<cdot> ?\<sigma>" .
  } note \<tau> = this
  {
    fix t :: "('f,'v)term"
    have "t \<cdot> \<tau> = t \<cdot> (\<lambda> x. \<mu> (T x) \<cdot> ?\<sigma>)" unfolding \<tau>[symmetric] ..
    also have "... = map_vars T t \<cdot> \<mu> \<cdot> ?\<sigma>" unfolding map_subst_domain 
      subst_compose_apply_term_distrib[symmetric]
      by (rule term_subst_eq, simp add: subst_compose_def)
    finally have "t \<cdot> \<tau> = map_vars T t \<cdot> \<mu> \<cdot> ?\<sigma>" .
  } note idt = this
  {
    fix x 
    assume "x \<in> S"
    hence "\<sigma> x = ?\<sigma> x" by simp
    also have "... = (\<mu> \<circ>s ?\<sigma>) x" using \<sigma> by simp
    also have "... = \<mu> x \<cdot> ?\<sigma>" unfolding subst_compose_def ..
    finally have "\<sigma> x = \<mu> x \<cdot> ?\<sigma>" .
  } note \<sigma> = this
  show ?thesis 
    by (rule exI[of _ \<mu>], rule exI[of _ ?\<sigma>], insert \<mu> ids idt \<sigma>, auto)
qed

lemma mgu_var_disjoint_right_string: fixes s t :: "('f,string)term" and \<sigma> \<tau> :: "('f,string)subst" 
  assumes s: "vars_term s \<subseteq> range x_var \<union> range z_var"
  and id: "s \<cdot> \<sigma> = t \<cdot> \<tau>"
  shows "\<exists> \<mu> \<delta>. mgu s (map_vars y_var t) = Some \<mu> \<and> s \<cdot> \<sigma> = s \<cdot> \<mu> \<cdot> \<delta> \<and> (\<forall> t :: ('f,string)term. t \<cdot> \<tau> = map_vars y_var t \<cdot> \<mu> \<cdot> \<delta>) \<and> (\<forall>x \<in> range x_var \<union> range z_var. \<sigma> x = \<mu> x \<cdot> \<delta>)"
proof -
  have inj: "inj y_var" unfolding inj_on_def by simp
  show ?thesis 
    by (rule mgu_var_disjoint_right[OF s inj _ id], auto)
qed

end
