(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory XML_Parser
imports
  Monadic_Check
  Show_Instances
  Linear_Orders
begin

abbreviation tab :: "char" where "tab \<equiv> Char Nibble0 Nibble9"
abbreviation newline :: char where "newline \<equiv> Char Nibble0 NibbleA"
abbreviation carriage_return :: "char" where "carriage_return \<equiv> Char Nibble0 NibbleD"

datatype xml = XML string "(string \<times> string) list" "xml list" "string option" (* name, attributes, children, text *)
datatype xmldoc = XMLDOC "string" xml (* header, body *)

text {* An alternative induction scheme for xml. *}
lemma
 fixes P :: "xml \<Rightarrow> bool"
 assumes "\<And>n a cs t. (\<And>c. c \<in> set cs \<Longrightarrow> P c) \<Longrightarrow> P(XML n a cs t)"
 shows xml_induct[case_names XML, induct type: xml]: "P xml" and "\<And>xml. xml \<in> set xmls \<Longrightarrow> P xml"
using assms by (induct xml and xmls) auto

text {* Resolve problem with recursive dependencies in generated code. *}
lemma [code]:
  shows "HOL.eq (XML name1 attrs1 children1 text1) (XML name2 attrs2 children2 text2) \<longleftrightarrow>
    HOL.eq name1 name2 \<and> HOL.eq attrs1 attrs2 \<and> list_all2 HOL.eq children1 children2 \<and> HOL.eq text1 text2"
  by (simp_all add: list_all2_eq [symmetric])

definition wspace :: "string" where [code_unfold]: "wspace = CHR '' '' # newline # tab # carriage_return # []"
definition dquote :: "char" where [code_unfold]: "dquote = CHR ''\"''"
definition letters :: "string" where "letters = ''abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789&;:''"

fun
  trim :: "string \<Rightarrow> string" 
where
  "trim (c # s) = (if c \<in> set wspace then trim s else c # s)"
| "trim [] = []"

lemma trim: "\<exists> w. s = w @ trim s"
proof (induct s)
  case (Cons c t)
  show ?case 
  proof (cases "c \<in> set wspace")
    case False 
    with Cons show ?thesis by auto
  next
    case True
    hence trim: "trim (c # t) = trim t" by auto
    from Cons obtain w where t: "t = w @ trim t" by auto
    show ?thesis 
      by (simp only: trim, rule exI[of _ "c # w"], simp, rule t)
  qed
qed simp

fun
  splitter_fun :: "(char \<Rightarrow> bool) \<Rightarrow> string \<Rightarrow> string \<times> string"
where
  "splitter_fun allowed (c # s) = (if allowed c
     then (let (l,r) = splitter_fun allowed s in (c # l, r))
     else ([], c # s))"
| "splitter_fun _ [] = ([],[])"

definition
  splitter :: "string \<Rightarrow> string \<Rightarrow> string \<times> string"
where
  [code_unfold]: "splitter allowed s = splitter_fun (\<lambda>c. c \<in> set allowed) s"

type_synonym  'a reply = "string + 'a"
type_synonym  ('t, 'a) gen_parser = "'t list \<Rightarrow> ('a \<times> 't list) reply"
type_synonym  'a parser = "(char, 'a) gen_parser"

subsection {* Setup parsers as monads *}

abbreviation (input)
  return :: "'a \<Rightarrow> ('t, 'a) gen_parser"
where
  "return x \<equiv> (\<lambda>ts. Error_Monad.return (x, ts))"

abbreviation (input)
  error :: "string \<Rightarrow> ('t, 'a) gen_parser"
where
  "error e \<equiv> (\<lambda>_. Error_Monad.error e)"

definition
  bind :: "('t, 'a) gen_parser \<Rightarrow> ('a \<Rightarrow> ('t, 'b) gen_parser) \<Rightarrow> ('t, 'b) gen_parser"
where
  "bind m f ts = do {
     (x, ts') \<leftarrow> m ts;
     f x ts'
   }"


setup {*
  Adhoc_Overloading.add_variant @{const_name Monad_Syntax.bind} @{const_name bind}
*}

lemma bind_cong[fundef_cong]:
 fixes m1 :: "('t,'a)gen_parser"
 assumes "m1 ts2 = m2 ts2"
 and "\<And> y ts. m2 ts2 = Inr (y,ts) \<Longrightarrow> f1 y ts = f2 y ts"
 and "ts1 = ts2"
 shows "((m1 \<guillemotright>= f1) ts1) = ((m2 \<guillemotright>= f2) ts2)"
using assms unfolding bind_def by (cases "m1 ts1", auto)


(* types 'a endParser = "string \<Rightarrow> (string,'a)emonad" *)

definition
  update_tokens :: "('t list \<Rightarrow> 't list) \<Rightarrow> ('t, 't list) gen_parser"
where
  "update_tokens f ts \<equiv> Error_Monad.return (ts, f ts)"

definition
  get_tokens :: "('t, 't list) gen_parser"
where
  "get_tokens ts \<equiv> Error_Monad.return (ts, ts)"

definition
  set_tokens :: "'t list \<Rightarrow> ('t, unit) gen_parser"
where
  [code_unfold]: "set_tokens ts \<equiv> update_tokens (\<lambda>_. ts) \<guillemotright> return ()"

definition
  expecting :: "string \<Rightarrow> ('t::show, 'a) gen_parser"
where
  "expecting msg ts = Error_Monad.error
     (''expecting: '' @ msg @ '', but found: '' @ shows_quote (shows (take 30 ts)) [])"

fun
  eoi :: "('t::show, unit) gen_parser"
where
  "eoi [] = Error_Monad.return ((), [])" |
  "eoi ts = expecting ''end of input'' ts"

fun exactly_aux :: "string \<Rightarrow> string \<Rightarrow> string \<Rightarrow> string parser" where
  "exactly_aux s i (x # xs) (y # ys) = (if x = y then exactly_aux s i xs ys else expecting (''exactly ['' @ s @ '']'') i)"
| "exactly_aux s i [] xs = Error_Monad.return (s,trim xs)"
| "exactly_aux s i (x # xs) [] = expecting (''exactly ['' @ s @ '']'') i"

fun
  oneof_aux :: "string list \<Rightarrow> string list \<Rightarrow> string parser"
where
  "oneof_aux allowed (x # xs) ts = (if map snd (zip x ts) = x
     then Error_Monad.return (x, trim (List.drop (length x) ts))
     else oneof_aux allowed xs ts)"
| "oneof_aux allowed [] ts = expecting (''one of '' @ shows_list allowed []) ts"

definition parser :: "'a parser \<Rightarrow> bool"
  where "parser p \<equiv> \<forall> s r x. p s = Inr (x,r) \<longrightarrow> length s \<ge> length r"

definition cparser :: "'a parser \<Rightarrow> bool"
  where "cparser p \<equiv> \<forall> s r x. p s = Inr (x,r) \<longrightarrow> length s > length r"

lemma parser_parser: assumes p: "parser p" and q: "\<And> x. parser (q x)" 
  shows "parser (do { x \<leftarrow> p; q x})"
unfolding parser_def
proof (intro allI impI)
  fix s r x
  assume "(p \<guillemotright>= q) s = Inr (x,r)"
  then obtain y t where P: "p s = Inr (y, t)" and Q: "q y t = Inr (x,r)" unfolding bind_def 
    by (cases "p s", auto)
  have "length r \<le> length t" using q[unfolded parser_def] Q by auto
  also have "\<dots>   \<le> length s" using p[unfolded parser_def] P by auto
  finally show "length r \<le> length s" .
qed
  
definition
  oneof :: "string list \<Rightarrow> string parser"
where
  "oneof xs \<equiv> oneof_aux xs xs"

lemma oneof_result: assumes "oneof xs s = Inr (y,r)" 
  shows "\<exists> w. s = y @ w @ r \<and> y \<in> set xs"
using assms
unfolding oneof_def
proof -
  assume ass: "oneof_aux xs xs s = Inr (y, r)"
  { 
    fix ys
    assume "oneof_aux ys xs s = Inr (y,r)"
    hence "\<exists> w. s = y @ w @ r \<and> y \<in> set xs"
    proof (induct xs)
      case Nil thus ?case by (simp add: expecting_def)
    next
      case (Cons z zs)
      thus ?case 
      proof (cases "map snd (zip z s) = z")
        case False with Cons show ?thesis by simp
      next
        case True
        hence s: "s = z @ drop (length z) s"          
        proof (induct z arbitrary: s)
          case (Cons a zz)
          thus ?case 
            by (cases s, auto)
        qed simp
        from trim[of "drop (length z) s"] obtain w where "drop (length z) s = w @ trim (drop (length z) s)" by blast
        with s have s: "s = z @ w @ trim (drop (length z) s)" by auto
        from True Cons(2) have yz: "y = z" and r: "r = trim (drop (length y) s)" by auto
        show ?thesis 
          by (simp add: yz r, rule exI, rule s)
      qed
    qed
  }  
  from this[OF ass]
  show ?thesis .
qed

definition exactly :: "string \<Rightarrow> string parser" where 
  "exactly s x = exactly_aux s x s x"

lemma exactly_result: assumes "exactly x s = Inr (y,r)" 
  shows "\<exists> w. s = x @ w @ r \<and> y = x"
using assms
unfolding exactly_def
proof -
  assume ass: "exactly_aux x s x s = Inr (y, r)"
  { 
    fix a b
    assume "exactly_aux a b x s = Inr (y,r)"
    hence "\<exists> w. s = x @ w @ r \<and> y = a"
    proof (induct x arbitrary: s)
      case Nil 
      thus ?case using trim[of s] by auto
    next
      case (Cons c xs) note xs = this
      show ?case 
      proof (cases s)
        case Nil
        with xs show ?thesis by (auto simp: expecting_def)
      next
        case (Cons d ss)
        note xs = xs[unfolded Cons]
        from xs(2) have "exactly_aux a b xs ss = Inr (y,r) \<and> c = d" by (cases "c = d", auto simp: expecting_def)
        hence res: "exactly_aux a b xs ss = Inr (y,r)" and c: "c = d" by auto
        from xs(1)[OF res]
        show ?thesis unfolding Cons c by auto
      qed
    qed
  }
  from this[OF ass] show ?thesis .
qed

hide_const oneof_aux exactly_aux
 
lemma oneof_length: 
  assumes "oneof xs s = Inr (y, r)"
  shows "length s \<ge> length y + length r \<and> y \<in> set xs"
proof -
  from oneof_result[OF assms]
  obtain w where "s = y @ w @ r \<and> y \<in> set xs" ..
  thus ?thesis  by auto
qed

lemma oneof_parser: "parser (oneof ts)"
unfolding parser_def
proof (intro allI impI)
  fix s r x
  assume "oneof ts s = Inr (x ,r)"
  from oneof_length[OF this] show " length s \<ge> length r" by auto
qed

lemma oneof_cparser: assumes "\<forall> x \<in> set ts. length x \<ge> 1"
  shows "cparser (oneof ts)"
unfolding cparser_def
proof (intro allI impI)
  fix s r x
  assume "oneof ts s = Inr (x ,r)"
  from oneof_length[OF this] assms  show " length s > length r" by auto
qed

lemma exactly_length: 
  assumes "exactly x s = Inr (y, r)"
  shows "length s \<ge> length x + length r"
proof -
  from exactly_result[OF assms]
  obtain w where "s = x @ w @ r" by auto
  thus ?thesis  by auto
qed

lemma exactly_parser: "parser (exactly xs)"
unfolding parser_def
proof (intro allI impI)
  fix s r x
  assume "exactly xs s = Inr (x ,r)"
  from exactly_length[OF this] show " length s \<ge> length r" by auto
qed

lemma exactly_cparser: assumes "length xs \<ge> 1"
  shows "cparser (exactly xs)"
unfolding cparser_def
proof (intro allI impI)
  fix s r x
  assume "exactly xs s = Inr (x ,r)"
  from exactly_length[OF this] show " length s > length r" using assms by auto
qed

fun
  many :: "(char \<Rightarrow> bool) \<Rightarrow> (char list) parser"
where
  "many P (t # ts) = (if P t
     then do {
       (rs, ts') \<leftarrow> many P ts;
       Error_Monad.return (t # rs, ts')
     } else Error_Monad.return ([], t # ts))"
| "many P [] = Error_Monad.return ([], [])"

lemma many_parser: "parser (many P)" 
unfolding parser_def
proof (intro allI impI)
  fix s r x
  assume "many P s = Inr (x,r)"
  thus "length r \<le> length s"
  proof (induct s arbitrary: x r, simp)
    case (Cons t ts)
    thus ?case by (cases "P t", cases "many P ts", auto, force)
  qed
qed

definition
  manyof :: "char list \<Rightarrow> (char list) parser"
where
  [code_unfold]: "manyof cs \<equiv> many (\<lambda>c. c \<in> set cs)"

lemma manyof_parser: "parser (manyof cs)"
unfolding manyof_def by (rule many_parser)

definition
  spaces :: "unit parser"
where
  [code_unfold]: "spaces \<equiv> manyof wspace \<guillemotright> return ()"

lemma return_parser: "parser (return x)" unfolding parser_def by auto
lemma error_parser: "parser (error x)" unfolding parser_def by auto
lemma if_parser: assumes p: "parser p" and q: "parser q" shows 
  "parser (if b then p else q)"
using assms unfolding parser_def by (cases b, auto)

lemma spaces_parser: "parser spaces"
unfolding spaces_def by (rule parser_parser[OF manyof_parser return_parser])

fun
  scan_upto :: "string \<Rightarrow> string parser"
where
  "scan_upto end (t # ts) = (if map snd (zip end (t # ts)) = end
     then do {
       Error_Monad.return (end, List.drop (length end) (t # ts))
     } else do {
       (res, ts') \<leftarrow> scan_upto end ts;
       Error_Monad.return (t # res, ts')
     })"
| "scan_upto end [] = Error_Monad.error (''did not find end-marker '' @ shows_quote (shows end) [])"

lemma scan_upto_length: 
  assumes "scan_upto end s = Inr (y, r)"
  shows "length s \<ge> length end + length r"
using assms
proof (induct s arbitrary: y r)
  case (Cons t ts)
  show ?case 
  proof (cases "map snd (zip end (t # ts)) = end")
    case True
    then obtain tss where tss: "tss = t # ts" and map: "map snd (zip end tss) = end" by auto
    from map have len: "length tss \<ge> length end"
    proof (induct "end" arbitrary: tss)
      case (Cons e en)
      thus ?case by (cases tss, auto)
    qed simp
    from True tss Cons(2) have y: "y = end" and r: "r = List.drop (length end) tss" by auto
    show ?thesis by (simp only: tss[symmetric], simp add: y r, auto simp: len)
  next
    case False
    with Cons obtain res ts' where "scan_upto end ts = Inr (res,ts')" by (cases "scan_upto end ts", auto)
    from Cons(1)[OF this] Cons(2) False this show ?thesis by auto
  qed
qed simp

lemma scan_upto_parser: "parser (scan_upto end)"
unfolding parser_def using scan_upto_length[of "end"] by force

lemma scan_upto_cparser: "cparser (scan_upto (e # end))"
unfolding cparser_def using scan_upto_length[of "e # end"] by force

definition many_letters where [simp]: "many_letters \<equiv> manyof letters"

definition
  parse_name :: "string parser"
where
  "parse_name = do {
     n \<leftarrow> many_letters;
     spaces;
     (if n = [] 
       then error letters
       else return n)
   }"

abbreviation letters_impl :: "char \<Rightarrow> bool" where 
  "letters_impl c \<equiv> c \<ge> CHR ''a'' \<and> c \<le> CHR ''z'' \<or>
   c \<ge> CHR ''A'' \<and> c \<le> CHR ''Z'' \<or>
   c \<ge> CHR ''0'' \<and> c \<le> CHR ''9'' \<or>
   c \<in> set ''_&;:''"

lemma letters_impl [code]: "(c \<in> set letters) = letters_impl c"
proof (cases c)
  case (Char n1 n2)
  show ?thesis unfolding Char letters_def
    by (cases n1, simp_all add: nat_of_char.simps, (cases n2, simp_all)+)
qed

lemma many_letters [code_unfold]: "many_letters = many letters_impl"
  unfolding letters_impl[symmetric] 
  by (simp add: manyof_def)

lemma parse_name: "parser parse_name" 
unfolding parse_name_def many_letters_def
by (rule parser_parser[OF manyof_parser parser_parser[OF spaces_parser if_parser[OF error_parser return_parser]]])

definition
  parse_text :: "string option parser" 
where
  "parse_text = do {
     ts \<leftarrow> many (op \<noteq> (CHR ''<''));
     let text = trim ts;
     (if text = []
       then return None
       else return (Some (List.rev (trim (List.rev text)))))
   }"

lemma parse_text: "parser parse_text"
unfolding parse_text_def Let_def
by (rule parser_parser[OF many_parser if_parser[OF return_parser return_parser]])

definition
  parse_attribute_value :: "string parser"
where
  "parse_attribute_value = do {
     exactly [dquote];
     v \<leftarrow> many (op \<noteq> dquote);
     exactly [dquote];
     return v
   }"


lemma parse_attribute_value: "parser parse_attribute_value"
  by (unfold parse_attribute_value_def, 
    rule parser_parser[OF exactly_parser parser_parser[OF many_parser parser_parser[OF exactly_parser return_parser]]]) 


function (sequential)
  parse_attributes :: "(string \<times> string) list parser"
where
  "parse_attributes [] = Error_Monad.return ([], [])" |
  "parse_attributes (c # s) = (if c \<in> set ''/>''
     then Error_Monad.return ([], c # s)
     else ((do {
       k \<leftarrow> parse_name;
       exactly ''='';
       v \<leftarrow> parse_attribute_value;
       atts \<leftarrow> parse_attributes;
       return ((k, v) # atts)
     }) (c # s)))"
by (pat_completeness) auto

termination parse_attributes
proof
  show "wf (measures [length])" by simp
next
  fix c s y ts ya tsa yb tsb
  assume pn: "parse_name (c # s) = Inr (y, ts)"
  and oo: "exactly ''='' ts = Inr (ya,tsa)"
  and pav: "parse_attribute_value tsa = Inr (yb, tsb)"  
  have cp: "cparser (exactly ''='')" by (rule exactly_cparser, auto)
  from cp[unfolded cparser_def] oo have 1: "length ts > length tsa" by auto
  from parse_name[unfolded parser_def] pn have 2: "length (c # s) \<ge> length ts" by force
  from parse_attribute_value[unfolded parser_def] pav have 3: "length tsa \<ge> length tsb" by force
  from 1 2 3
  show "(tsb, c # s) \<in> measures[length]" 
    by auto
qed

lemma parse_attributes: "parser parse_attributes"
unfolding parser_def
proof (intro allI impI)
  fix s r x 
  assume "parse_attributes s = Inr (x,r)"
  thus "length r \<le> length s"
  proof (induct arbitrary: x rule: parse_attributes.induct , simp)
    case (2 c s)
    show ?case
    proof (cases "c \<in> set ''/>''")
      case True
      with 2(2) show ?thesis by simp
    next
      case False
      from False 2(2) obtain y1 s1 where pn: "parse_name (c # s) = Inr (y1, s1)" by (cases "parse_name (c # s)", auto simp: bind_def)
      from False 2(2) pn obtain y2 s2 where oo: "exactly ''='' s1 = Inr (y2, s2)" by (cases "exactly ''='' s1", auto simp: bind_def)
      from False 2(2) pn oo obtain y3 s3 where pav: "parse_attribute_value s2 = Inr (y3, s3)" by (cases "parse_attribute_value s2", auto simp: bind_def)
      from False 2(2) pn oo pav obtain y4 s4 where patts: "parse_attributes s3 = Inr (y4, r)" by (cases "parse_attributes s3", auto simp: bind_def)
      have "length r \<le> length s3" using 2(1)[OF False pn oo pav patts] .
      also have "\<dots> \<le> length s2" using parse_attribute_value[unfolded parser_def] pav by auto
      also have "\<dots> \<le> length s1" using exactly_parser[unfolded parser_def] oo by auto
      also have "\<dots> \<le> length (c # s)" using parse_name[unfolded parser_def] pn by force
      finally show "length r \<le> length (c # s)" by auto
    qed
  qed
qed

function parse_nodes where  "parse_nodes ts = 
     (if ts = [] \<or> take 2 ts = ''</''
       then return [] ts
       else (do {
         exactly ''<'';
         n \<leftarrow> parse_name;
         atts \<leftarrow> parse_attributes;
         e \<leftarrow> oneof [''/>'', ''>''];
         (\<lambda> ts'. if (e = ''/>'') then
           (do {
             cs \<leftarrow> parse_nodes;
             return (XML n atts [] None # cs)
           }) ts' else 
           (do {
             t \<leftarrow> parse_text;
             cs \<leftarrow> parse_nodes;
             exactly ''</'';
             exactly n;
             exactly ''>'';
             ns \<leftarrow> parse_nodes;
             return (XML n atts cs t # ns)
           }) ts')
       }) ts)"
by (pat_completeness) auto

lemma parse_nodes_help: "parse_nodes_dom s \<and> (\<forall> x r. parse_nodes s = Inr (x,r) \<longrightarrow> length r \<le> length s)" (is "?prop s")
proof (induct rule: wf_induct[where P = ?prop and r = "measures [ length]"])
  fix s
  assume "\<forall> t. (t, s) \<in> measures [length] \<longrightarrow> ?prop t"
  hence ind1: "\<And> t. length t < length s \<Longrightarrow> parse_nodes_dom t"  
    and ind2: "\<And> t x r. length t < length s \<Longrightarrow> parse_nodes t = Inr (x,r) \<Longrightarrow> length r \<le> length t" by auto
  let ?check = "\<lambda> s. s = [] \<or> take 2 s = ''</''"
  have dom: "parse_nodes_dom s"
  proof
    fix y 
    assume "parse_nodes_rel y s"
    thus "parse_nodes_dom y"
    proof
      fix ts ya tsa yaa tsb yb tsc yc tsd 
      assume "y = tsd" and "s = ts" and "\<not> ?check ts" and "exactly ''<'' ts = Inr (ya, tsa)" and "parse_name tsa = Inr (yaa, tsb)"
        and "parse_attributes tsb = Inr (yb, tsc)" and "oneof [''/>'', ''>''] tsc = Inr (yc, tsd)" and "yc = ''/>''"
      hence len: "length s > length y" using exactly_cparser[of "''<''", simplified] oneof_parser[of "[''/>'', ''>'']"] parse_attributes parse_name
        unfolding parser_def cparser_def by force
      with ind1[OF len] show "parse_nodes_dom y" by simp
    next
      fix ts ya tsa yaa tsb yb tsc yc tsd yd tse
      assume "y = tse" and "s = ts" and "\<not> ?check ts" and "exactly ''<'' ts = Inr (ya, tsa)" and "parse_name tsa = Inr (yaa, tsb)"
        and "parse_attributes tsb = Inr (yb, tsc)" and "oneof [''/>'', ''>''] tsc = Inr (yc, tsd)" and "parse_text tsd = Inr (yd, tse)"
      hence len: "length s > length y" using exactly_cparser[of "''<''", simplified] oneof_parser[of "[''/>'', ''>'']"] parse_attributes parse_name parse_text
        unfolding parser_def cparser_def by force
      with ind1[OF len] show "parse_nodes_dom y" by simp
    next
      fix ts ya tsa yaa tsb yb tsc yc tsd yd tse ye tsf yf tsg yg tsh yh tsi yi tsj
      assume y: "y = tsj" and "s = ts" and "\<not> ?check ts" and "exactly ''<'' ts = Inr (ya, tsa)" and "parse_name tsa = Inr (yaa, tsb)"
        and "parse_attributes tsb = Inr (yb, tsc)" and "oneof [''/>'', ''>''] tsc = Inr (yc, tsd)" and "parse_text tsd = Inr (yd, tse)"
        and rec: "parse_nodes_sumC tse = Inr (ye, tsf)" 
        and last: "exactly ''</'' tsf = Inr (yf, tsg)" "exactly yaa tsg = Inr (yg, tsh)" "exactly ''>'' tsh = Inr (yi, tsj)"
      hence len: "length s > length tse" using exactly_cparser[of "''<''", simplified] oneof_parser[of "[''/>'', ''>'']"]  parse_attributes parse_name parse_text
        unfolding parser_def cparser_def by force
      from last(1) last(2) have len2a: "length tsf \<ge> length tsh" using exactly_parser[of "''</''"] exactly_parser[of yaa] parse_name unfolding parser_def by force
      have len2c: "length tsh \<ge> length y" using last(3) 
        using exactly_parser[of "''>''"] unfolding parser_def y by simp
      from len2a len2c have len2: "length tsf \<ge> length y" by simp
      from ind2[OF len rec[unfolded parse_nodes_def[symmetric]]] len len2 have "length s > length y" by simp
      from ind1[OF this]       
      show "parse_nodes_dom y" .
    qed
  qed
  note psimps = parse_nodes.psimps[OF dom]
  show "?prop s"
  proof (intro conjI, rule dom, intro allI impI)
    fix x r    
    assume res: "parse_nodes s = Inr (x,r)"
    note res = res[unfolded psimps]
    thus "length r \<le> length s"
    proof (cases "?check s")
      case True      
      thus ?thesis using res by simp
    next
      case False note oFalse = this
      note res = res[simplified False, simplified]
      from res obtain y1 s1 where oo: "exactly ''<'' s = Inr (y1, s1)" by (cases "exactly ''<'' s", auto simp: bind_def)
      note res = res[unfolded bind_def oo, simplified]
      from res obtain y2 s2 where pn: "parse_name s1 = Inr (y2, s2)" by (cases "parse_name s1", auto simp: bind_def psimps)
      note res = res[simplified bind_def pn, simplified]
      from res obtain y3 s3 where pa: "parse_attributes s2 = Inr (y3, s3)" by (cases "parse_attributes s2", auto)
      note res = res[simplified pa, simplified]
      from res obtain y4 s4 where oo2: "oneof [''/>'', ''>''] s3 = Inr (y4, s4)" by (cases "oneof [''/>'', ''>''] s3", auto)
      note res = res[unfolded oo2, simplified]
      from parse_attributes oneof_parser[of "[''/>'', ''>'']"] exactly_cparser[of "''<''", simplified] parse_name
        oo pn pa oo2
      have s_s4: "length s > length s4" unfolding parser_def cparser_def by force
      show ?thesis
      proof (cases "y4 = ''/>''")
        case True
        from res True obtain y5 where pns: "parse_nodes s4 = Inr (y5, r)" by (cases "parse_nodes s4", auto)
        from ind2[OF s_s4 pns] s_s4 show "length r \<le> length s"  by simp
      next
        case False
        from res False obtain y5 s5 where pt: "parse_text s4 = Inr (y5, s5)" by (cases "parse_text s4", auto)
        note res = res[simplified False, simplified, unfolded pt, simplified,
          unfolded bind_def]
        from res obtain y6 s6 where pns: "parse_nodes s5 = Inr (y6, s6)" by (cases "parse_nodes s5", auto)
        note res = res[unfolded bind_def pns, simplified, unfolded bind_def]
        from res obtain y7 s7 where oo3: "exactly ''</'' s6 = Inr (y7, s7)" by (cases "exactly ''</'' s6", auto)
        note res = res[unfolded oo3, simplified, unfolded bind_def, 
          simplified, unfolded bind_def]
        from res obtain y8 s8 where oo4: "exactly y2 s7 = Inr (y8, s8)" by (cases "exactly y2 s7", auto)
        note res = res[unfolded oo4 bind_def, simplified]
        from res obtain y10 s10 where oo5: "exactly ''>'' s8 = Inr (y10,s10)"
          by (cases "exactly ''>'' s8", auto simp: bind_def)
        note res = res[unfolded oo5 bind_def, simplified]
        from res obtain y11 s11 where pns2: "parse_nodes s10 = Inr (y11, s11)" by (cases "parse_nodes s10", auto simp: bind_def)
        note res = res[unfolded bind_def pns2, simplified]
        from pt parse_text s_s4 have s_s5: "length s > length s5" unfolding parser_def by force
        note one = oneof_parser[unfolded parser_def, rule_format]
        note exact = exactly_parser[unfolded parser_def, rule_format]
        from ind2[OF s_s5 pns] s_s5 exact[OF oo3] exact[OF oo4] have s_s7: "length s > length s8" unfolding parser_def by force 
        with exact[OF oo5] have s_s10: "length s > length s10" by simp
        with ind2[OF s_s10 pns2] have s_s11: "length s > length s11" by simp
        thus "length r \<le> length s" using res by auto
      qed
    qed
  qed
qed simp

termination parse_nodes using parse_nodes_help by blast

lemma parse_nodes: "parser parse_nodes" 
  unfolding parser_def using parse_nodes_help by blast 

(* more efficient version of oneof [''/>'', ''>''] *)
fun oneof_closed :: "string parser" where
  "oneof_closed (x # xs) = (if x = CHR ''>'' then Error_Monad.return (''>'',trim xs)
     else if (x = CHR ''/'' \<and> (case xs of [] \<Rightarrow> False | Cons y ys \<Rightarrow> y = CHR ''>'')) then Error_Monad.return (''/>'', trim (tl xs))
     else expecting (''one of [/>,>]'') (x # xs))"
| "oneof_closed xs = expecting (''one of [/>,>]'') xs"

lemma oneof_closed: "oneof [''/>'', ''>''] = oneof_closed" (is "?l = ?r")
proof (rule ext)
  fix xs  
  have id: "''one of '' @ shows_list [''/>'',''>''] [] = ''one of [/>,>]''" by eval
  note d = oneof_def oneof_aux.simps id
  show "?l xs = ?r xs"
  proof (cases xs)
    case Nil
    show ?thesis unfolding Nil d by simp
  next
    case (Cons x xs) note oCons = this
    show ?thesis
    proof (cases "x = CHR ''>''")
      case True
      show ?thesis unfolding Cons d True by simp
    next
      case False note oFalse = this
      show ?thesis
      proof (cases "x = CHR ''/''")
        case False
        show ?thesis unfolding Cons d using False oFalse by simp
      next
        case True
        show ?thesis
        proof (cases xs)
          case Nil
          show ?thesis unfolding Cons Nil d by auto
        next
          case (Cons y ys)
          show ?thesis unfolding oCons Cons d by simp
        qed
      qed
    qed
  qed
qed

lemma parse_if_removal: "(\<lambda> e x. if b e then f e x else g e x) = (\<lambda> e. if b e then f e else g e)"
  by (intro ext, auto)

declare parse_nodes.simps[unfolded oneof_closed, unfolded parse_if_removal[of "\<lambda> e. e = ''/>''"], code]

definition
  parse_node :: "xml parser"
where
  "parse_node \<equiv> do {
     exactly ''<'';
     n \<leftarrow> parse_name;
     atts \<leftarrow> parse_attributes;
     e \<leftarrow> oneof [''/>'', ''>''];
     (if e = ''/>''
       then return (XML n atts [] None)
       else do {
         t \<leftarrow> parse_text;
         cs \<leftarrow> parse_nodes;
         exactly ''</'';
         exactly n;
         exactly ''>'';
         return (XML n atts cs t)
       })
   }"

declare parse_node_def[unfolded oneof_closed, code]

function
  parse_header :: "string parser"
where
  "parse_header ts = (
     (if take 2 (trim ts) = ''<?''
       then (do {
         h \<leftarrow> scan_upto ''?>'';
         hs \<leftarrow> parse_header;
         return (h @ hs)
       }) ts else (do {
         spaces;
         return []
       }) ts)
   )"
by (pat_completeness) auto

termination parse_header
proof
  fix ts y tsa
  assume "scan_upto ''?>'' ts = Inr (y, tsa)"
  with scan_upto_cparser have "length ts > length tsa" unfolding cparser_def by force
  thus "(tsa, ts) \<in> measures [length]" by simp
qed simp

consts scala :: bool
code_const scala
  (Haskell "False")
  (Scala "true")
  (SML "false")
  (OCaml "false")
  (Eval "false")

fun
  remove_comments_aux :: "bool \<Rightarrow> string \<Rightarrow> string"
where
  "remove_comments_aux False (c # cs) = (if (c = CHR ''<'' \<and> take 3 cs = ''!--'') 
     then remove_comments_aux True (tl cs)
     else c # remove_comments_aux False cs)" |
  "remove_comments_aux True (c # cs) = (if (c = CHR ''-'' \<and> take 2 cs = ''->'') 
     then remove_comments_aux False (drop 2 cs)
     else remove_comments_aux True cs)" |
  "remove_comments_aux _ [] = []"

fun
  remove_comments_aux_acc :: "string \<Rightarrow> bool \<Rightarrow> string \<Rightarrow> string"
where
  "remove_comments_aux_acc a False (c # cs) = (if (c = CHR ''<'' \<and> take 3 cs = ''!--'') 
     then remove_comments_aux_acc a True (tl cs)
     else remove_comments_aux_acc (c # a) False cs)" |
  "remove_comments_aux_acc a True (c # cs) = (if (c = CHR ''-'' \<and> take 2 cs = ''->'') 
     then remove_comments_aux_acc a False (drop 2 cs)
     else remove_comments_aux_acc a True cs)" |
  "remove_comments_aux_acc a _ [] = a"

(* tail recursive variant for Scala to reduce Stack-size at the cost of double traversion *)
definition
  remove_comments :: "string \<Rightarrow> string"
where
  "remove_comments \<equiv> if scala then rev o (remove_comments_aux_acc [] False) else remove_comments_aux False"

hide_const remove_comments_aux remove_comments_aux_acc

definition
  parse_doc :: "xmldoc parser"
where
  "parse_doc = debug ''0'' ''parsing txt to xml'' (do {
     update_tokens remove_comments;
     h \<leftarrow> parse_header;
     xml \<leftarrow> parse_node;
     eoi;
     return (XMLDOC h xml)
   })"

definition
  doc_of_string :: "string \<Rightarrow> string + xmldoc"
where
  "doc_of_string s = do {
     (doc, _) \<leftarrow> parse_doc s;
     Error_Monad.return doc
   }"
  
end
